//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.command;

import com.nexirius.util.assertion.Assert;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

/**
 * This type was created in VisualAge
 *
 * @author Marcel Baumann
 * @invariant
 */
public class ActionCommandFactory {




    private static class ActionCommand
            extends AbstractAction
            implements PropertyChangeListener {
        private Processor processor;
        private Command command;

        private javax.swing.JButton button = null;

        private ActionCommand(Processor processor, Command command) {
            Assert.pre(processor != null, "Parameter processor is not null");
            Assert.pre(command != null, "Parameter command is not null");
            this.processor = processor;
            this.command = command;

            // Make abstract action listener of command property changes
            command.addPropertyChangeListener(this);

            // Set action name to be the same as the command
            putValue(Action.NAME, command.getCommandName());
        }

        private ActionCommand(Command command) {
            Assert.pre(command != null, "Parameter command is not null");
            this.command = command;
            this.processor = null;

            // Make abstract action listener of command property changes
            command.addPropertyChangeListener(this);

            // Set action name to be the same as the command
            putValue(Action.NAME, command.getCommandName());
        }

        private ActionCommand(Command command, javax.swing.JButton button) {
            this(command);

            this.button = button;
        }

        public void actionPerformed(ActionEvent e) {
            if (processor != null) {
                // Give process the command using the processor
                processor.execute(command, e);
            } else {
                // Process the command directly
                command.initialiseAction(e);
                command.doAction();
                command.finaliseAction();
            }
        }

        public void propertyChange(PropertyChangeEvent e) {
            if (e.getSource() != this.command) {
                return;
            }

            // Pass on the command property change to the abstract action
            putValue(e.getPropertyName(), e.getNewValue());

            if (e.getPropertyName().equals("enabled")) {
                boolean enabled = ((Boolean) e.getNewValue()).booleanValue();
                System.out.println(enabled);
                setEnabled(enabled);

                if (this.button != null) {
                    this.button.setEnabled(enabled);
                }
            }
        }
    }

    /**
     * Factory method returning a Swing Action who's actionPerformed()
     * will execute the command on the given processor
     *
     * @param command Command
     * @author Marcel Baumann
     */
    public static Action create(Processor processor, Command command) {
        return new ActionCommand(processor, command);
    }

    /**
     * Factory method returning a Swing Action who's actionPerformed()
     * will execute the command directly (i.e. without using a processor)
     *
     * @param command Command
     * @author Marcel Baumann
     */
    public static Action create(Command command) {
        return new ActionCommand(command);
    }


    public static Action create(Command command, javax.swing.JButton button) {
        return new ActionCommand(command, button);
    }


    public static void main(String[] args) {
        class TestCommand extends com.nexirius.framework.command.AbstractCommand {
            public void initialiseAction(Object param) {
                System.out.println("Command initialised");
            }

            public boolean doAction() {
                System.out.println("Command done");
                return(true);
            }

            public void finaliseAction() {
                System.out.println("Command finalised");
            }
        }

        javax.swing.JFrame f = new javax.swing.JFrame("Test Frame");
        f.addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent e) {
                System.out.println("Test program finished.");
                System.exit(0);
            }
        });

        final javax.swing.JButton b = new javax.swing.JButton("Press Me");

        final Command c = new TestCommand();
        final Action a = ActionCommandFactory.create(c, b);

        javax.swing.JButton tb = new javax.swing.JButton("Enable / Disable 'Press Me' Button");
        tb.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                // Test to see if the widget itself is enabled/disabled
                // in response to this setting:
                c.setEnabled(!c.isEnabled());
                System.out.println("Enabled? " + c.isEnabled());
            }
        });

        javax.swing.JMenuBar menuBar = new javax.swing.JMenuBar();
        javax.swing.JMenu menu = new javax.swing.JMenu("File");
        menuBar.add(menu);
        menu.add(a);

        b.addActionListener(a);

        final javax.swing.JToolBar toolBar = new javax.swing.JToolBar();
        toolBar.add(a);

        f.setJMenuBar(menuBar);
        f.getContentPane().setLayout(new java.awt.BorderLayout());
        f.getContentPane().add(b, java.awt.BorderLayout.CENTER);
        f.getContentPane().add(tb, java.awt.BorderLayout.SOUTH);
        f.getContentPane().add(toolBar, java.awt.BorderLayout.NORTH);

        f.setSize(400, 400);
        f.setVisible(true);
    }
}
