//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.command;

import com.nexirius.util.assertion.Assert;

/**
 * Event objects of this class are fired by context managers in response to
 * changes in context. There are two basic event types: REMOVE_CONTEXT and
 * ADD_CONTEXT. These correspond to whether a context was removed or added
 * from/to a context manager.
 *
 * @author Marcel Baumann
 * @version Date        Chng.Rq.   Author    Changes/Enhancements
 *          1999.01.04             MB       Created
 *          1999.03.03             MB       Required enhancement: Include info about
 *          what has changed in the context
 */
public class ContextChangeEvent extends java.util.EventObject {



    /**
     * Type constant indicating the removal of a context from a context manager
     * Listeners can query what was added via getRemovedContext
     *
     * @see #getChangeType()
     */
    public static final int REMOVE_CONTEXT = 1;

    /**
     * Type constant indicating the addition of a context to a context manager.
     * Listeners can query what was added via getAddedContext
     *
     * @see #getChangeType()
     */
    public static final int ADD_CONTEXT = 2;

    /**
     * Type constant indicating that something has changed about the context
     * which is neither an addition or a removal of a context
     * Hence listeners should requery the context using getContext
     *
     * @see #getChangeType()
     */
    public static final int SOME_CHANGE = 99;

    // The context manager which fired the event
    private ContextManager contextManager;

    // The context which this event concerns
    private Context context;

    // The context which this event concerns.
    // Only set if getChangeType() == ADD_CONTEXT
    private Context addedContext = null;

    // The context which this event concerns.
    // Only set if getChangeType() == REMOVE_CONTEXT
    private Context removedContext = null;

    // The type of context event
    private int contextChangeType;

    /**
     * Construct a context change event
     *
     * @param contextManager    The object which fired this event
     * @param context           The context which which event concerns
     * @param contextChangeType The type of the event. Note: Must not be
     *                          ADD_CONTEXT or REMOVE_CONTEXT. For these types, use other constructor
     */
    public ContextChangeEvent(ContextManager contextManager, Context context,
                              int contextChangeType) {
        super(contextManager);

        Assert.pre(contextManager != null,
                "Paramater contextManager is not null");
        Assert.pre(context != null, "Paramater context is not null");
        Assert.pre(contextChangeType != ADD_CONTEXT &&
                contextChangeType != REMOVE_CONTEXT,
                "Parameter contextChangeType " +
                " is neither ADD_CONTEXT nor REMOVE_CONTEXT");

        this.contextManager = contextManager;
        this.context = context;
        this.contextChangeType = contextChangeType;
    }


    /**
     * Construct a context change event
     *
     * @param contextManager        The object which fired this event
     * @param context               The context which which event concerns
     * @param addedOrRemovedContext The context which was added or removed
     * @param contextChangeType     The type of the event
     */
    public ContextChangeEvent(ContextManager contextManager, Context context,
                              Context addedOrRemovedContext, int contextChangeType) {
        super(contextManager);

        Assert.pre(contextManager != null,
                "Paramater contextManager is not null");
        Assert.pre(context != null, "Paramater context is not null");
        Assert.pre(addedOrRemovedContext != null,
                "Parameter addedOrRemovedContext is not null");
        Assert.pre(contextChangeType == ADD_CONTEXT ||
                contextChangeType == REMOVE_CONTEXT,
                "Parameter contextChangeType " +
                " is either ADD_CONTEXT or REMOVE_CONTEXT");

        this.contextManager = contextManager;
        this.context = context;
        switch (contextChangeType) {
            case ADD_CONTEXT:
                {
                    addedContext = addedOrRemovedContext;
                    break;
                }
            case REMOVE_CONTEXT:
                {
                    removedContext = addedOrRemovedContext;
                    break;
                }
        }
        this.contextChangeType = contextChangeType;
    }


    /**
     * Return the context manager which fired this event
     *
     * @return the context manager which caused this event to be fired
     * @author Marcel Baumann
     */
    public ContextManager getContextManager() {
        return this.contextManager;
    }


    /**
     * Return the complete context. This generally represents all the commands
     * currently available.
     *
     * @return the context representing all the currently available commands
     * @author Marcel Baumann
     */
    public Context getContext() {
        return this.context;
    }


    /**
     * Return the context which was added
     * Note: only if getType() == ADD_CONTEXT
     *
     * @return the context which was added
     * @author Marcel Baumann
     */
    public Context getAddedContext() {
        Assert.pre(getChangeType() == ADD_CONTEXT,
                "change type is ADD_CONTEXT");
        return this.addedContext;
    }


    /**
     * Return the context which was removed
     * Note: only if getType() == REMOVE_CONTEXT
     *
     * @return the context which was removed
     * @author Marcel Baumann
     */
    public Context getRemovedContext() {
        Assert.pre(getChangeType() == REMOVE_CONTEXT,
                "change type is REMOVE_CONTEXT");
        return this.removedContext;
    }


    /**
     * Return the type of the event
     *
     * @return int
     * @author Marcel Baumann
     */
    public int getChangeType() {
        return contextChangeType;
    }
}
