//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.command;

import com.nexirius.util.assertion.Assert;

import java.util.Stack;

/**
 * ContextStackManager provides a mechanism to add and remove groups of
 * commands at runtime as the state of the program changes (i.e. Context
 * changes).  The idea is to push basic application commands (exit,
 * cut-copy-paste, help) onto the bottom of the stack at start up, then
 * and add and remove commands at runtime as the context changes.  For
 * example if a text editor window appears, it can override the basic
 * cut-copy-paste commands by pushing a Context object onto the stack.
 * Likewise new commands can be added specific to the editor.  When the
 * editor window is closed, its context objects should be removed as well.
 * <p/>
 * TODO!!
 * 1) Do we need a way to add a Context into the middle of the stack?
 * 2) There is no way to get back the individual Context in which a command
 * resides.  The whole stack is treated as a single context (not efficient).
 * 3) TODO!! Duplicate commands are not removed in getCommandList().
 *
 * @author EYI
 * @date 1999.01.04
 * @see Context
 */
public class ContextStackManager implements ContextManager, Context {



    private transient ContextChangeListener aContextChangeListener = null;

    // This class maintains a stack of commands
    private Stack stack = new Stack();

    /**
     * Construct a stack context manager object
     *
     * @param event com.nexirius.framework.CommandSetEvent
     * @author EYI
     */
    public ContextStackManager() {
    }

    ////////////////////////////////////////////////////////////////////////////
    //  ContextManager interface
    ////////////////////////////////////////////////////////////////////////////

    /**
     * Return a context object which represents all of the commands in the
     * current context i.e. this!
     *
     * @author EYI
     */
    public Context getContext() {
        return this;
    }


    /**
     * Add a context to the context manager, by pushing it on top
     * of the stack.
     *
     * @param context the context to be added
     * @author EYI
     */
    public synchronized void addContext(Context context) {
        Assert.pre(context != null, "Parameter context is not null");
        if (stack.contains(context)) {
            return;
        }
        Context c = (Context) stack.push(context);
        fireContextChange(context, ContextChangeEvent.ADD_CONTEXT);
    }


    /**
     * Remove a context from the context manager
     *
     * @param context the context to be removed
     * @author EYI
     */
    public synchronized void removeContext(Context context) {
        Assert.pre(context != null, "Parameter context is not null");
        if (stack.removeElement(context)) {
            fireContextChange(context, ContextChangeEvent.REMOVE_CONTEXT);
        }
    }


    /**
     * Add a context change listener
     *
     * @param listner context change listener
     * @author EYI
     */
    public void addContextChangeListener(ContextChangeListener listener) {
        aContextChangeListener =
                ContextChangeEventMulticaster.add(aContextChangeListener, listener);
    }


    /**
     * Remove a context change listener
     *
     * @param listener remove a context change listener
     * @author EYI
     */
    public void removeContextChangeListener(ContextChangeListener listener) {
        aContextChangeListener =
                ContextChangeEventMulticaster.remove(aContextChangeListener, listener);
    }


    /**
     * Method to support listener events
     *
     * @param event com.nexirius.framework.CommandSetEvent
     * @author EYI
     */
    private void fireContextChange(Context addedOrRemovedContext, int contextChangeType) {
        if (aContextChangeListener == null) {
            return;
        }
        aContextChangeListener.contextChange(new ContextChangeEvent(this, this,
                addedOrRemovedContext, contextChangeType));
    }


    ////////////////////////////////////////////////////////////////////////////
    //  Context interface
    ////////////////////////////////////////////////////////////////////////////

    /**
     * Search for the presence of a commandName in the Context stack
     * starting from the top of the stack down.  If no match is found
     * returns false.
     *
     * @param commandName java.lang.String
     * @return boolean
     * @author EYI
     */
    public boolean containsCommand(String commandName) {
        Assert.pre(commandName != null, "Parameter commandName is not null");

        for (int i = stack.size() - 1; i >= 0; i--) {
            Context c = (Context) stack.elementAt(i);
            if (c.containsCommand(commandName))
                return true;
        }
        return false;
    }


    /**
     * Return the command assocated with the given command name starting
     * from the top of the Context stack.  If two commands with identical
     * names are present in the stack, the upper instance is returned.
     * An assertion is thrown if the commandName is not present (i.e.
     * call containsCommand() first!!).
     *
     * @param commandName java.lang.String
     * @return com.nexirius.framework.command.Command
     * @author EYI
     */
    public Command getCommand(String commandName) {
        Assert.pre(commandName != null, "Parameter commandName is not null");
        Assert.pre(containsCommand(commandName), "Context contains command");

        for (int i = stack.size() - 1; i >= 0; i--) {
            Context c = (Context) stack.elementAt(i);
            if (c.containsCommand(commandName)) {
                return c.getCommand(commandName);
            }
        }

        Assert.assertion(false, "getCommand() always returns a command");
        return null;
    }

    /**
     * Return a complete list of all the commands in this context.
     * In other words, treats the stack as if it was one big single context.
     * TODO!! Duplicate commands are not removed presenlty!!
     *
     * @return com.nexirius.framework.command.Command[]
     * @author EYI
     */
    public Command[] getCommandList() {
        int count = 0;
        for (int i = 0; i < stack.size(); i++) {
            Context c = (Context) stack.elementAt(i);
            Command[] cmds = c.getCommandList();
            count += cmds.length;
        }
        int index = 0;
        Command[] cmdList = new Command[count];
        for (int i = 0; i < stack.size(); i++) {
            Context c = (Context) stack.elementAt(i);
            Command[] cmds = c.getCommandList();
            for (int j = 0; j < cmds.length; j++) {
                cmdList[index++] = cmds[j];
            }
        }
        return cmdList;
    }


    /**
     * Return the name of this context
     *
     * @return java.lang.String
     * @author EYI
     */
    public String getName() {
        return getClass().getName();
    }


    /**
     * Return the processor associated with the given command
     *
     * @return com.nexirius.framework.command.Processor
     * @author EYI
     */
    public Processor getProcessor(String commandName) {
        Assert.pre(commandName != null, "Parameter commandName is not null");
        Assert.pre(containsCommand(commandName),
                "Context contains command " + commandName);

        for (int i = stack.size() - 1; i >= 0; i--) {
            Context c = (Context) stack.elementAt(i);
            if (c.containsCommand(commandName)) {
                return c.getProcessor(commandName);
            }
        }

        // This should never happen, hence
        Assert.assertion(false, "Processor always be returned from getProcessor()");
        return null;
    }
}
