//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.command;

import com.nexirius.util.assertion.Assert;

/**
 * This type represents a basic implementation of the Processor interface.
 * Commands are executed by calling the execute() method. The command is
 * executed depending on the commands execution property e.g. timeout,
 * synchronous etc.
 * <p/>
 * !! This default implementation of the processor *does not* support undo/redo
 * although this should in theory be quite easy to add.
 *
 * @author Marcel Baumann
 */
public class DefaultProcessor implements Processor {
    // Inner class used to execute a single command in another thread
    private class CommandExecutor extends Thread {
        // The command to be executed
        private Command command;

        // The parameter to be passed to initialiseAction
        private Object param = null;

        // Flag to indicate command completion
        private boolean completed = false;

        /**
         * Construct a command executor
         *
         * @param command The command to be executed
         * @param param   The paramater to be passed to the command's
         *                initialiseAction() method
         */
        CommandExecutor(Command command, Object param) {
            this.command = command;
            this.param = param;
        }

        /**
         * Perform the actial execution of the command
         *
         * @see java.lang.Runnable#run()
         */
        public void run() {
            try {
                this.command.initialiseAction(param);
                this.command.doAction();
                this.command.finaliseAction();
            } catch (Throwable e) {
                // Invoke the exception action on the command
                e.printStackTrace();
                this.command.exceptionAction(e);
            }

            // Signal to the outside world that the command completed
            this.completed = true;
        }

        /**
         * Return whether or not command execution completed
         *
         * @return boolean
         */
        public boolean isCompleted() {
            return this.completed;
        }
    }


    /**
     * Construct a default command processor
     */
    public DefaultProcessor() {
        super();
    }


    /**
     * Execute the command
     *
     * @param command The command to be executed
     * @param param   The paramater to be passed to the command's
     *                initialiseAction() method.
     * @author Marcel Baumann
     */
    public void execute(final Command command, Object param) {
        Assert.pre(command != null, "Parameter command is not null");

        // Reset abort after timeout flag
        command.setAbortAfterTimeout(false);

        final CommandExecutor commandExecutor = new CommandExecutor(command, param);

        if (command.isSynchronous()) {
            commandExecutor.run();
        } else {
            final Thread t = new Thread(commandExecutor, "Asynchronous " + command.getCommandName());

            t.start();

            if (command.getTimeout() > 0) {
                new Thread(new Runnable() {
                    public void run() {
                        try {
                            Thread.sleep(command.getTimeout());

                            if (!commandExecutor.isCompleted()) {
                                // Give command opportunity to react
                                commandExecutor.command.timeoutAction();

                                // Find out if command wants to continue
                                if (commandExecutor.command.getAbortAfterTimeout()) {
                                    t.interrupt();
                                }
                            }
                        } catch (InterruptedException e) {
                            // should not happen
                        }
                    }
                }, "Watching asynchronous " + command.getCommandName()).start();
            }
        }
    }
}
