//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 *
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 *
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 *
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 *
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.dataeditor;


import com.nexirius.framework.datamodel.*;
import com.nexirius.framework.dataviewer.DataViewer;
import com.nexirius.framework.dataviewer.LayoutItem;
import com.nexirius.framework.dataviewer.ViewerCreator;
import com.nexirius.util.assertion.Assert;

import javax.swing.*;
import javax.swing.border.BevelBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.awt.*;

/**
 * This editor is designed to edit ArrayModel data. The default component
 * is a single vertical column of all the children editors. The associated
 * layout is used to layout the array items
 *
 * @author Marcel Baumann
 */
public class ArrayEditor extends DataViewer {



    public static final String editorName = "ArrayEditor";

    StepperModel stepper;
    Updater updater = null;
    int maxConcurrentElements = 1;
    ViewerCreator elementCreator = null;

    /**
     * Creates a new struct viewer
     */
    public ArrayEditor(ArrayModel model, ViewerCreator elementCreator, int maxConcurrentElements) {
        super(model);
        this.maxConcurrentElements = maxConcurrentElements;
        this.elementCreator = elementCreator;
    }

    /**
     * Returns true
     */
    @Override
    public boolean isEditor() {
        return true;
    }

    public void setElementCreator(ViewerCreator elementCreator) {
        this.elementCreator = elementCreator;
    }

    /**
     * Creates a field value component (data model editor for a specific child)
     *
     * @param dm     The child data model which will be used to create an editor component.
     * @param layout the layout used to create the editor (or null)
     */
    public JComponent createFieldValue(DataModel dm, LayoutItem layout)
            throws Exception {
        DataViewer v = null;

        if (elementCreator == null) {
            v = getFactory().createDefaultEditor(dm);
            ;
        } else {
            v = elementCreator.createViewer(dm);
            v.setFactory(getFactory());
        }

        v.setLayout(layout);

        JComponent ret = v.getJComponent();

        ret.setName(dm.getFieldName());

        return ret;
    }

    /**
     * Only for debugging
     */
    @Override
    public String getViewerName() {
        return editorName;
    }

    /**
     * Returns the associated model
     */
    public ArrayModel getModel() {
        return (ArrayModel) getDataModel();
    }

    /**
     * Returns the assotiated JPanel or null (if called before creation)
     */
    public JPanel getJPanel() {
        if (isCreated()) {
            return (JPanel) ((JScrollPane) getJComponent().getComponent(1)).getViewport().getView();
        }

        return null;
    }

    /**
     * Returns the assotiated JPanel or null (if called before creation)
     */
    public JPanel getStepperPanel() {
        if (isCreated()) {
            return (JPanel) getJComponent().getComponent(0);
        }

        return null;
    }

    /**
     * Creates the actual JPanel and all the children components (depends on the asssociated layout status)
     */
    @Override
    public void create() {
        Assert.pre(this.factory != null, "Can't create a StructViewer with null factory (call setFactory())");

        JScrollPane sp = new JScrollPane(new JPanel(new GridBagLayout()), JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        sp.setFocusable(false);
        sp.getViewport().setFocusable(false);

        sp.getViewport().addChangeListener(new MyChangeListener());

        JPanel mainPanel = new JPanel(new GridBagLayout());

        mainPanel.add(new JPanel(new GridLayout(1, 1)), new GridBagConstraints(0, 0, 1, 1, 1.0, 0.0, GridBagConstraints.NORTH, GridBagConstraints.HORIZONTAL, new Insets(0, 0, 0, 0), 0, 0));
        mainPanel.add(sp, new GridBagConstraints(0, 1, 1, 1, 1.0, 1.0, GridBagConstraints.NORTH, GridBagConstraints.BOTH, new Insets(0, 0, 0, 0), 0, 0));

        setJComponent(mainPanel);

        update();
    }

    /**
     * this class handles the down sizing in width
     */
    class MyChangeListener implements ChangeListener {
        int oldWidth = -1;

        @Override
        public void stateChanged(ChangeEvent e) {
            JViewport viewPort = (JViewport) e.getSource();

            int width = viewPort.getWidth();

            if (oldWidth < 0 || oldWidth > width) {
                update(width);
            }

            oldWidth = width;
        }
    }

    /**
     * Not used
     */
    @Override
    public void dataModelChangeValue(DataModelEvent event) {
        if (getDataModel() == event.getDataModel()) {
            update();
        } else {
            revalidate();
        }

        if (isCreated() && getDataModel() == event.getDataModel() && getDataModel().isInvisible()) {
            setVisible(!getDataModel().isInvisible());
        }
    }

    public void setVisible(boolean on) {
        if (!isCreated()) {

            return;
        }

        if (isVisible() != on) {
            SwingUtilities.invokeLater(new SetVisible(on));
        }
    }

    public boolean isVisible() {
        if (!isCreated()) {

            return false;
        }

        return getJComponent().isVisible();
    }

    class SetVisible implements Runnable {
        boolean on;

        SetVisible(boolean _on) {
            on = _on;
        }

        @Override
        public void run() {
            getJComponent().setVisible(on);
        }
    }

    /**
     * Calls update
     *
     * @see #update
     */
    @Override
    public void dataModelChangeStructure(DataModelEvent event) {
    }

    public void revalidate() {
        if (isCreated()) {
            getJPanel().revalidate();
        }
    }

    @Override
    public void update() {
    }

    public synchronized void update(int width) {
        if (!isCreated()) {

            return;
        }

        if (updater != null) {
            updater.interrupt();
        }

        updater = new Updater(width);

        updater.start();
    }

    class Updater extends Thread {
        private int width;

        Updater(int width) {
            this.width = width;
        }

        @Override
        public void run() {
            /**
             Removes all components of the current JPanel and recreates them. This method is
             automatically called whenever the associated data model structure fires a structure
             change event.
             */
            if (isCreated()) {
                if (stepper == null) {
                    stepper = new StepperModel(0, getModel().getSize(), "Stepper");

                    stepper.setCount(maxConcurrentElements);

                    stepper.addUpdateListener(new StepperUpdateListener());
                    getModel().addDataModelListener(new ArraySelectionListener());

                    try {
                        DataViewer stepperViewer = getFactory().createDefaultEditor(stepper);
                        getStepperPanel().add(stepperViewer.getJComponent());
                    } catch (Exception ex) {
                        //FIX
                        throw new RuntimeException(ex);
                    }
                }

                GridBagConstraints gridbag = new GridBagConstraints(0, 0, 1, 1, 1.0, 0.0, GridBagConstraints.NORTH, GridBagConstraints.HORIZONTAL, new Insets(0, 0, 0, 0), 0, 0);

                if (width > 0) {
                    //getJPanel().setMaximumSize(new Dimension(width, Integer.MAX_VALUE));
                }

                getJPanel().removeAll();
                getJPanel().repaint();
                getJPanel().setLayout(new GridBagLayout());

                stepper.setMaxCount(getModel().getSize());
                int i = 0;
                int gridy = 0;

                for (DataModel dm = getModel().getChildren().firstItem(); dm != null && i <= stepper.getTo(); dm = getModel().getChildren().nextItem()) {

                    if (isInterrupted()) {
                        //FIX System.out.println("STOPPED");

                        return;
                    }

                    if (i >= stepper.getFrom()) {
                        try {
                            JPanel panel = new JPanel(new BorderLayout());
                            JComponent comp = createFieldValue(dm, getLayout());

                            if (comp != null) {
                                panel.add(comp, BorderLayout.CENTER);
                                gridbag.gridy = gridy;
                                getJPanel().add(panel, gridbag);
                                panel.setBorder(new BevelBorder(BevelBorder.RAISED));
                                ++gridy;
                            }
                        } catch (Exception ex) {
                            //FIX
                            throw new RuntimeException(ex);
                        }
                    }

                    ++i;
                }

                getJPanel().add(new JPanel(), new GridBagConstraints(0, gridy, 1, 1, 1.0, 1.0, GridBagConstraints.NORTH, GridBagConstraints.BOTH, new Insets(0, 0, 0, 0), 0, 0));
                getJPanel().revalidate();

                updateUI();
            }
        }
    }

    /**
     * Not implemented for JPanel
     */
    @Override
    public void grabFocus() {
    }

    class StepperUpdateListener extends DataModelAdaptor {
        @Override
        public void dataModelChangeValue(DataModelEvent ev) {
            getModel().setHighlightedItem(stepper.getFrom());
            update();
        }
    }

    class ArraySelectionListener extends DataModelAdaptor {
        @Override
        public void dataModelEdit(DataModelEvent ev) {
            if (ev.getId() == DataModelEvent.HIGHLIGHT_CHANGED) {
                if (stepper != null) {

                    if (getModel().getHighlightedItemIndex() >= 0) {
                        stepper.setFrom(getModel().getHighlightedItemIndex());
                    }
                }
            }
        }
    }
}
