//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 *
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 *
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 *
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 *
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.datamodel;

import com.nexirius.util.CopyPairs;
import com.nexirius.util.simpletype.SimpleType_BigDecimal;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.DecimalFormat;
import java.text.ParseException;
import java.text.ParsePosition;

/**
 * A class that represents a BigDecimal information which can be displayed in a text field.
 *
 * @author Marcel Baumann
 */
public class BigDecimalModel extends SimpleModel {
    public static ModelFlag INACTIVE_FLAG = new ModelFlag("INACTIVE");
    protected BigDecimal maxValue = null;
    protected BigDecimal minValue = null;
    protected BigDecimalFormat bigDecimalFormat;
    protected RoundingMode roundingMode = null;
    protected int scale;

    public BigDecimalModel() {
        this(BigDecimal.ZERO);
    }

    /**
     * Creates a new instance and initializes its value
     *
     * @param value The initial value
     */
    public BigDecimalModel(BigDecimal value) {
        this(value, null);
    }

    /**
     * Creates a new instance and initializes its value and field name
     *
     * @param value     The initial value
     * @param fieldName The initial field name
     */
    public BigDecimalModel(BigDecimal value, String fieldName) {
        super(new SimpleType_BigDecimal(value), fieldName);

        if (value == null) {
            setFlag(ModelFlag.EMPTY, true);
        }
    }

    @Override
    public void setNotEmpty(boolean notEmpty) {
        super.setNotEmpty(notEmpty);

        if (notEmpty && isEmpty()) {
            setBigDecimal(BigDecimal.ZERO);
        }
    }

    public boolean isActive() {
        return !getFlag(INACTIVE_FLAG);
    }

    public boolean setActive(boolean on) {
        return setFlag(INACTIVE_FLAG, !on);
    }

    /**
     * sets the rounding mode for this big decimal
     *
     * @param scale
     * @param roundingMode
     */
    public void setRoundingMode(int scale, RoundingMode roundingMode) {
        this.scale = scale;
        this.roundingMode = roundingMode;
    }

    /**
     * change the value to a rounded big decimal (see BigDecimal. setScale())
     * call setRoundingMode() first otherwise this method has no effect
     */
    public void round() {
        if (roundingMode != null) {
            setBigDecimal(roundValue(getBigDecimal()));
        }
    }

    public BigDecimal roundValue(BigDecimal value) {
        if (roundingMode != null) {
            return value.setScale(scale, roundingMode);
        }

        return value;
    }

    public void setBigDecimalFormat(BigDecimalFormat bigDecimalFormat) {
        this.bigDecimalFormat = bigDecimalFormat;
    }

    /**
     * factor i
     *
     * @param format
     * @param factor Factor is used to multiply the actual value before using the actual decimal format pattern (may be null)
     */
    public void setBigDecimalFormat(String format, BigDecimal factor) {
        BigDecimalFormat f = new BigDecimalFormat(format, factor);
        setBigDecimalFormat(f);
    }

    /**
     * Returns the current value as BigDecimal or BigDecimal.ZERO if isEmpty()
     */
    public BigDecimal getBigDecimal() {
        if (isEmpty()) {
            return BigDecimal.ZERO;
        }
        return getSimpleType().getBigDecimal();
    }

    /**
     * Set the new value as BigDecimal (by default the value is not rounded!)
     *
     * @param v The new value
     */
    public void setBigDecimal(BigDecimal v) {
        setBigDecimal(v, false);
    }

    /**
     * Set the new value as BigDecimal
     *
     * @param v The new value
     * @param doRound round the value before setting it
     */
    public void setBigDecimal(BigDecimal v, boolean doRound) {
        if (v == null && notEmpty) {
            v = BigDecimal.ZERO;
        }

        if (v == null) {
            if (!isEmpty()) {
                value = new SimpleType_BigDecimal(BigDecimal.ZERO);
                setFlag(ModelFlag.EMPTY, true);
            }
        } else {
            if (doRound) {
                v = roundValue(v);
            }

            boolean forceValueChangeEvent = false;

            if (isEmpty()) {
                // use getStatus to avoid firing flag changed event
                getStatus().setFlag(ModelFlag.EMPTY, false);
                forceValueChangeEvent = true;
            }

            setValue(new SimpleType_BigDecimal(v), forceValueChangeEvent);
        }
    }

    public BigDecimal getMinValue() {
        return minValue;
    }

    public void setMinValue(BigDecimal minValue) {
        this.minValue = minValue;
    }

    public BigDecimal getMaxValue() {
        return maxValue;
    }

    public void setMaxValue(BigDecimal maxValue) {
        this.maxValue = maxValue;
    }

    public void setZero() {
        setBigDecimal(BigDecimal.ZERO);
    }

    public boolean isZero() {
        return getBigDecimal().compareTo(BigDecimal.ZERO) == 0;
    }

    public boolean isBiggerThan(BigDecimal v) {
        return getBigDecimal().compareTo(BigDecimal.ZERO) > 0;
    }

    public boolean isSmallerThan(BigDecimal v) {
        return getBigDecimal().compareTo(BigDecimal.ZERO) < 0;
    }

    public boolean isPositiveOrZero() {
        return getBigDecimal().compareTo(BigDecimal.ZERO) >= 0;
    }

    public boolean isNegative() {
        return getBigDecimal().compareTo(BigDecimal.ZERO) < 0;
    }

    public BigDecimal addValue(BigDecimal value) {
        BigDecimal result = getBigDecimal().add(value);

        setBigDecimal(result);

        return getBigDecimal();
    }

    public BigDecimal multiplyValue(BigDecimal value) {
        BigDecimal result = getBigDecimal().multiply(value);

        setBigDecimal(result);

        return getBigDecimal();
    }

    /**
     * Set the new value as string
     *
     * @param text The string which is parsed to produce the new integer value
     * @throws Exception If the text cannot be parsed into an integer.
     */
    @Override
    public void setText(String text) throws Exception {
        if (text == null || text.trim().length() == 0) {
            setBigDecimal(null);
        } else {
            BigDecimal newValue = null;

            if (bigDecimalFormat == null) {
                newValue = new BigDecimal(text);
            } else {
                try {
                    text = text.trim();
                    ParsePosition p = new ParsePosition(0);
                    newValue = bigDecimalFormat.parse(text.trim(), p);

                    if (p.getIndex() != text.length()) {
                        throw new Exception("parse failed");
                    }
                } catch (Exception ex) {
                    newValue = new BigDecimal(text);
                }
            }

            checkLimits(newValue);
            setBigDecimal(newValue);
        }
    }

    public void checkLimits(BigDecimal newValue) throws Exception {
        if (newValue == null) {
            return;
        }

        if (maxValue != null && maxValue.compareTo(newValue) < 0) {
            throw new Exception("bigger than maxValue");
        }

        if (minValue != null && minValue.compareTo(newValue) > 0) {
            throw new Exception("smaller than minValue");
        }
    }

    /**
     * Returns an exact copy of this instance
     *
     * @param instance  null (only set by subclasses dupicate method)
     * @param copyPairs null (only needed with container classes)
     */
    @Override
    public DataModel duplicate(DataModel instance, CopyPairs<DataModel> copyPairs) {
        if (instance == null) {
            BigDecimalModel copy = new BigDecimalModel(getBigDecimal());
            copy.bigDecimalFormat = bigDecimalFormat;
            instance = copy;
        }

        return super.duplicate(instance, copyPairs);
    }

    /**
     * Returns the string representation of the current value (eventually using the current
     * number format)
     */
    @Override
    public String getText() {
        if (bigDecimalFormat == null || isExceptional()) {
            return super.getText();
        }

        return bigDecimalFormat.format(getBigDecimal());
    }

    public String format(BigDecimal value) {
        if (value == null) {
            return "";
        }

        if (bigDecimalFormat == null) {
            return value.toString();
        }

        return bigDecimalFormat.format(value);
    }

    public static class BigDecimalFormat {
        DecimalFormat decimalFormat;
        BigDecimal factor;

        public BigDecimalFormat(String pattern) {
            this(pattern, null);
        }

        public BigDecimalFormat(String pattern, BigDecimal factor) {
            this.decimalFormat = new DecimalFormat(pattern);
            this.decimalFormat.setParseBigDecimal(true);
            this.factor = factor;
        }

        public String format(BigDecimal value) {
            if (value == null) {
                return "";
            }

            if (factor != null) {
                value = value.multiply(factor);
            }

            return decimalFormat.format(value);
        }

        public BigDecimal parse(String text, ParsePosition parsePosition) throws ParseException {
            BigDecimal result = (BigDecimal) decimalFormat.parse(text, parsePosition);

            if (factor != null) {
                result = result.divide(factor);
            }

            return result;
        }
    }
}
