//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.datamodel;

import com.nexirius.util.CopyPairs;
import com.nexirius.util.TextToken;
import com.nexirius.util.simpletype.SimpleType_Date;

import java.io.OutputStream;
import java.io.PushbackInputStream;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * A class that represents a date information which can be displayed in a text field or in
 * a specific date viewer or editor (not yet implemented).
 * Instances of this class can only be used in editors or viewers if they have a DateFormat
 *
 * @author Marcel Baumann
 * @see #setDateFormat
 */
public class DateModel extends SimpleModel {
    public static final long NULL_TIME = -1L;

    public DateFormat defaultDateFormat = new SimpleDateFormat("dd. MMM yyyy HH:mm");
    private TimeModel timeModel;

    /**
     * Creates a new instance and initializes its value to 1.1.1970
     */
    public DateModel() {
        this(new Date(0));
        setEmpty();
    }

    /**
     * Creates a new instance and initializes its value to date.
     *
     * @param value The initial date value
     */
    public DateModel(Date value) {
        super(new SimpleType_Date(value));
        setDateFormat(defaultDateFormat);
    }

    /**
     * Creates a new instance and initializes its value and field name.
     *
     * @param value     The initial date value
     * @param fieldName The initial field name
     */
    public DateModel(Date value, String fieldName) {
        super(new SimpleType_Date(value), fieldName);
        setDateFormat(defaultDateFormat);
    }

    /**
     * Returns the current value as simple type
     */
    public SimpleType_Date getSimpleType_Date() {
        return (SimpleType_Date) getSimpleType();
    }

    /**
     * Set the date format which is used to parse text from test field input and to create a string from a date.
     *
     * @param dateFormat The new date format
     * @see java.text.SimpleDateFormat
     */
    public void setDateFormat(DateFormat dateFormat) {
        getSimpleType_Date().setDateFormat(dateFormat);
    }
   
    /**
     * Get the date format which should be used with this date.
     * @return the DateFormat associated with this date
     */
    public DateFormat getDateFormat() {
        return getSimpleType_Date().getDateFormat();
    }

    /**
     * Returns the current value as java.util.Date
     */
    public Date getDate() {
        return getSimpleType_Date().getDate();
    }

    /**
     * Set the value to a new date.
     *
     * @param date The new value (not null)
     */
    public void setDate(Date date) {
        if (isEmpty() && date != null) {
            setEmpty(false);
        }

        try {
            if (date == null) {
                date = new Date(0);
                setValue(new SimpleType_Date(date, getSimpleType_Date().getDateFormat()));
            } else if (date.getTime() != getDate().getTime()) {
                setValue(new SimpleType_Date(date, getSimpleType_Date().getDateFormat()));
            }
        } catch (Exception e) {
            throw new RuntimeException(e);  //To change body of catch statement use File | Settings | File Templates.
        }
    }

    /**
     * Uses the current DateFormat (@see setDateFormat) to parse the given string into a new date value.
     *
     * @param text The string which will be parsed to set the new date value.
     * @throws Exception If the given  string is no valid date string (from the date format point of view)
     */
    public void setText(String text)
            throws Exception {
        if (text == null || text.length() == 0) {
            setNull();
            setEmpty();
        } else if ("NOW".equalsIgnoreCase(text)) {
            setDate(new Date());
        } else {
            setValue(new SimpleType_Date(text, getSimpleType_Date().getDateFormat()));
            setEmpty(false);
        }
    }

    public long getTime() {
        return getDate().getTime();
    }

    public void setTime(long time) {
        if (time != getTime()) {
            getSimpleType_Date().setTime(time);

            if (!isEditing()) {
                fireValueChange();
            }
        }
    }

    public boolean isNull() {
        return getTime() == NULL_TIME;
    }

    public void setNull() {
        setDate(new Date(NULL_TIME));
    }


    /**
     * Creates an exact copy of this instance.
     */
    public DataModel duplicate(DataModel instance, CopyPairs copyPairs) {
        if (instance == null) {
            instance = new DateModel(getDate());
        }

        return super.duplicate(instance, copyPairs);
    }

    /**
     * Read persistent information from input stream
     *
     * @param in The input stream to read from
     * @throws Exception If the input stream is not readable or has corrupted data
     */
    public void readDataFrom(PushbackInputStream in)
            throws Exception {
        getStatus().readDataFrom(in, this);

        TextToken token = TextToken.nextToken(in);

        if (token.isString()) {
            String s = token.getString();

            setTime(Long.parseLong(s));
        } else {
            throw new Exception("Expecting String token but have:" + token);
        }
    }

    /**
     * Writes persistent information to output stream
     *
     * @param out The stream where to write to
     * @throws Exception If the output stream is not writable
     */
    public void writeDataTo(OutputStream out)
            throws Exception {
        getStatus().writeDataTo(out);

        TextToken value = new TextToken(Long.toString(getTime()));

        value.writeTo(out);
    }

    public TimeModel getTimeModel() {
        if (timeModel == null) {
            timeModel = new TimeModel(getDate(), "timeModel");

            timeModel.addDataModelListener(new DataModelAdaptor() {
                public void dataModelChangeValue(DataModelEvent event) {
                    setDate(timeModel.getDate(getDate()));
                }
            });
            addDataModelListener(new DataModelAdaptor() {
                public void dataModelChangeValue(DataModelEvent event) {
                    timeModel.setDate(getDate());
                }
            });
        }

        return timeModel;
    }
}
