//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.datamodel;

import com.nexirius.framework.command.AbstractCommand;
import com.nexirius.framework.application.ErrorMessageException;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * This class is the used to define a data model method (a command which is a member function to a
 * specific data model instance).
 * The DataModel method with the same name as the command (public and no parameters) will be invoked on doAction()
 * @author Marcel Baumann
 */
public class DefaultDataModelCommand extends AbstractCommand
        implements DataModelCommand {

    private static final long serialVersionUID = 7617249247560506303L;
    
    private boolean invokeAction = true;
    DataModel target = null;

    /**
     * Creates a command with the class name as command name.
     */
    public DefaultDataModelCommand() {
        super();
    }

    /**
     * Creates a command and initialises its name.
     *
     * @param name The command name
     */
    public DefaultDataModelCommand(String name) {
        super(name);
    }

    /**
     * Creates a command and initialises its name.
     *
     * @param name The command name
     * @param invokeAction If true (default) then the public DataModel method with no parameters will be invoked on doAction
     */
    public DefaultDataModelCommand(String name, boolean invokeAction) {
        super(name);
        this.invokeAction = invokeAction;
    }

    /**
     * return an exact copy
     */
    @Override
    public DataModelCommand duplicate() {
        DefaultDataModelCommand ret = new DefaultDataModelCommand(getCommandName());

        Hashtable<String, Object> properties = this.getProperties();
        Hashtable<String, Object> newProperties = ret.getProperties();
        Enumeration<String> e = properties.keys();

        while (e.hasMoreElements()) {
            String key = e.nextElement();

            newProperties.put(key, properties.get(key));
        }

        return ret;
    }

    /**
     * Associates a data model with this command.
     *
     * @param target Member function 'this' reference
     */
    @Override
    public void setDataModel(DataModel target) {
        this.target = target;
    }

    /**
     * Returns the associated DataModel instance ('this' reference)
     */
    @Override
    public DataModel getDataModel() {
        return target;
    }

    /**
     * Returns the field name of the command. This function is mainly used to create a button component (as resource key).
     *
     * @return By default the command name is returned
     * @see com.nexirius.framework.command.AbstractCommand#getCommandName
     */
    @Override
    public String getFieldName() {
        return getCommandName();
    }

    /**
     * This function should be redefined for a specific subclass which does something on the
     * associated data model. By default this function creates a debug information on the standard
     * output.
     * @return true if the action was successfull, false if anything unusal happened.
     */
    @Override
    public boolean doAction() {
//System.out.println("COMMAND: " + getCommandName() + " called");

        if (!invokeAction) 
        {
            // Sollte nicht passieren. Das doAction() müsste in einer Subklasse überschrieben werden.
            return (false);
        }

        try {
            Method meth[] = target.getClass().getMethods();
            boolean found = false;

            for (int i = 0; i < meth.length; ++i) {
                if (meth[i].getParameterTypes().length == 0 && meth[i].getName().equals(getCommandName())) {
                    meth[i].invoke(target, (Object[])null);
                    found = true;
                }
            }

            if (!found) {
                throw new Exception("METHOD " + getCommandName() + " not found in " + target.getClass().getName());
            } else {
            }
        } catch (InvocationTargetException ex) {
            if (ex.getTargetException() instanceof ErrorMessageException) {
                throw (ErrorMessageException)ex.getTargetException();
            }
            throw new RuntimeException("InvocationTargetException: " + ex.getTargetException().getClass().getName() + ": " + ex.getTargetException().getMessage(), ex);
        } catch (Exception ex) {
            throw new RuntimeException(ex.getClass().getName() + ": " + ex.getMessage(), ex);
        }
        return(true);
    }

    /**
     * Action to be performed in case an exception was thrown during command
     * execution
     *
     * @param e Throwable
     */
    /*@Override
    public void exceptionAction(Throwable e) {
        super.exceptionAction(e);
    }*/

    /**
     * This method is called before doAction. Override this method if something
     * should be performed in initialiseAction.
     * <p/>
     * The object passed as a parameter may used to pass initialising info
     * to the command.  If the command was invoked from a button or menu status
     * then the object may correspond to the action event (the application
     * programmer has to know).
     *
     * @param parameter free parameter
     */
    @Override
    public void initialiseAction(Object parameter) {
    }

    /**
     * This method is called after doAction. Override this method
     * if something should be performed in finaliseAction.
     * This method fires the CALL_METHOD DataModelEvent on the associated DataModel
     */
    @Override
    public void finaliseAction() {
        if (getDataModel() == null) {

            return;
        }

        DataModelEvent event = new DataModelEvent(getDataModel(), DataModelEvent.CALL_METHOD, true);

        event.setData(getCommandName());
        getDataModel().propagateChildEvent(event);
    }

    /**
     * This action is performed if a command times out.  Override this
     * if some specific action is required.
     * <p/>
     * The default behaviour is to write a message to the logger
     */
/*    @Override
    public void timeoutAction() {
        super.timeoutAction();
    }*/
}
