//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.datamodel;

import com.nexirius.util.CopyPairs;
import com.nexirius.util.simpletype.SimpleType_double;

import java.text.NumberFormat;
import java.text.ParsePosition;

/**
 * A class that represents an double information which can be displayed in a text field.
 *
 * @author Marcel Baumann
 */
public class DoubleModel extends SimpleModel {

    NumberFormat format = null;

    public DoubleModel() {
        this(0.0);
    }

    /**
     * Creates a new instance and initialises its value
     *
     * @param value The initial value.
     */
    public DoubleModel(double value) {
        super(new SimpleType_double(value));
    }

    /**
     * Creates a new instance and initialises its value and field name
     *
     * @param value     The initial value.
     * @param fieldName The initial field name
     */
    public DoubleModel(double value, String fieldName) {
        super(new SimpleType_double(value), fieldName);
    }

    /**
     * Creates a new instance and initialises its field name
     *
     * @param fieldName The initial field name
     */
    public DoubleModel(String fieldName) 
    {
       this(0.0, fieldName);
    }
    
    /**
     * Uses the NumberFormat class to translate and parse text input and output
     * <PRE>
     * myDouble.setNumberFormat(new java.text.DecimalFormat("#,##0.000' kg'"));
     * </PRE>
     *
     * @param format e.g. new DecimalFormat("'Fr.' #,##0.00")
     */
    public void setNumberFormat(NumberFormat format) {
        this.format = format;
    }

    /**
     * Get the current value as double
     */
    public double getDouble() {
        return getSimpleType().getDouble();
    }

    /**
     * Set the value as double
     */
    public void setDouble(double i) {
        setEmpty(false);
        setValue(new SimpleType_double(i));
    }

    /**
     * Set the value as string.
     * The current number format is used to parse the string into a double. If the parsing
     * fails then this method tries to parse the string as a plain floating point (using Double.valueOf())
     *
     * @throws Exception If the string is not parsed by the associated format and if it is no plain floating point number.
     */
    public void setText(String text)
            throws Exception {
        if (text.length() == 0) {
            if (!isEmpty()) {
                value = new SimpleType_double(0.0);
                setEmpty();
            }
        } else {
            if (isEmpty()) {
                setEmpty(false);
            }
            if (format == null) {
                setValue(new SimpleType_double(text));
            } else {
                try {
                    ParsePosition pos = new ParsePosition(0);

                    setValue(new SimpleType_double(format.parse(text, pos).doubleValue()));

                    if (pos.getIndex() != text.length()) {
                        throw new Exception();
                    }
                } catch (Exception ex) {
                    setValue(new SimpleType_double(text));
                }
            }
        }
    }

    /**
     * Returns the string representation of the current value (eventually using the current
     * number format)
     */
    public String getText() {
        if (format == null || isExceptional()) {
            return super.getText();
        }

        return format.format(getDouble()).toString();
    }

    /**
     * Creates an exact copy of the data model.
     *
     * @param instance  null (Only subclasses set this parameter to a specific value)
     * @param copyPairs null (Only container classes use this parameter to avoid multiple copies of the same instance)
     */
    @Override
    public DataModel duplicate(DataModel instance, CopyPairs<DataModel> copyPairs) {
        if (instance == null) {
            instance = new DoubleModel(getDouble());
        }

        ((DoubleModel) instance).setNumberFormat(format);

        return super.duplicate(instance, copyPairs);
    }
}
