//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.datamodel;

import com.nexirius.util.CopyPairs;

/**
 * A class that represents an string information which can be displayed in a text field.
 * It eventually disallows the input of more than a limited number of characters.
 *
 * @author Marcel Baumann
 */
public class LimitStringModel extends StringModel {



    int maxLength = -1;

    /**
     * @param value     The initial value
     * @param maxLength A maximum length or -1 (no length limitation)
     * @param fieldName The field name
     */
    public LimitStringModel(String value, int maxLength, String fieldName) {
        super(value, fieldName);
        setMaxLength(maxLength);
    }

    /**
     * Defines a string model with no length limitations
     *
     * @param value     The initial value
     * @param fieldName The field name
     */
    public LimitStringModel(String value, String fieldName) {
        super(value, fieldName);
        setMaxLength(-1);
    }

    /**
     * @param value     The initial value
     * @param maxLength A maximum length or -1 (no length limitation)
     */
    public LimitStringModel(String value, int maxLength) {
        super(value);
        setMaxLength(maxLength);
    }

    /**
     * Defines a string model with no length limitations
     *
     * @param value The initial value
     */
    public LimitStringModel(String value) {
        super(value);
        setMaxLength(-1);   // = no length limitation
    }

    /**
     * Set the maximum length of the current string value. If the current value is
     * longer than the new lenth the it automatically is truncated
     *
     * @param maxLength The new length limitation (-1 for unlimited length)
     * @see #validate
     */
    public void setMaxLength(int maxLength) {
        this.maxLength = maxLength;
        validate();
    }

    public int getMaxLength() {
        return maxLength;
    }

    /**
     * Returns true if the given string is shorter or equal in size than max length.
     * If maxLength is <0 then this method returns true.
     *
     * @param s This string value is tested against the current length settings.
     */
    public boolean isValid(String s) {
        return s == null || maxLength < 0 || s.length() <= maxLength;
    }

    /**
     * Returns the adjusted string if the given string is longer  in size than max length.
     * If maxLength is <0 then the same string instance is returned.
     *
     * @param s A substring of this string is returned if it exceeds the current max length.
     */
    public String makeValid(String s) {
        if (!isValid(s)) {
            return s.substring(0, maxLength);
        }

        return s;
    }

    /**
     * Eventually resets the current value so its length does not exeed the current limitation.
     */
    public void validate() {
        String s = makeValid(getText());

        if (!s.equals(getText())) {
            try {
                setText(s);
            } catch (Exception ex) {
                throw new RuntimeException(ex);
            }
        }
    }

    /**
     * Set the new value as string.
     *
     * @throws RuntimeException If the given text exceeds the current lenght limitations
     */
    public void setText(String text)
//		throws Exception
    {
        if (isValid(text)) {
            super.setText(text);
        } else {
            throw new RuntimeException("Value too long: '" + text + "' [max = " + maxLength + "]");
        }
    }

    /**
     * Returns an exact copy of this.
     *
     * @param instance  null (or a defined subclass instance when called from subclass method duplicate)
     * @param copyPairs null (not used)
     */
    public DataModel duplicate(DataModel instance, CopyPairs copyPairs) {
        if (instance == null) {
            instance = new LimitStringModel(getText(), maxLength);
        }

        return super.duplicate(instance, copyPairs);
    }
}
