package com.nexirius.framework.datamodel.json.format;

import com.nexirius.framework.datamodel.*;
import com.nexirius.framework.datamodel.json.*;
import com.nexirius.util.TextToken;
import com.nexirius.util.CopyPairs;

import java.io.PushbackInputStream;
import java.io.IOException;
import java.io.OutputStream;

public class ArrayJsonFormat implements IJsonFormat {
    private IJsonFormatMap jsonFormatMap;
    private DataModel template;

    public ArrayJsonFormat(IJsonFormatMap jsonFormatMap, DataModel template) {
        this.jsonFormatMap = jsonFormatMap;
        this.template = template;
    }

    public void parseJson(PushbackInputStream in, DataModel target) throws Exception, JsonParseException {
        ArrayModel array = (ArrayModel) target;
        TextToken token = TextToken.nextToken(in);

        if (token == null) {
            throw new JsonParseExceptionPrematureEndOfFile("in array " + array.getFieldName() + " before [");
        }

        if (!token.isChar('[')) {
            throw new JsonParseException(token, "[", "parsing array " + array.getFieldName());
        }

        CopyPairs pairs = new CopyPairs();

        while (true) {
            int c = in.read();

            if (c < 0) {
                throw new JsonParseExceptionPrematureEndOfFile("in array " + array.getFieldName() + " after [");
            }

            if (c == ']') {
                // we're done
                return;
            }

            if (c == ',' || Character.isWhitespace((char)c)) {
                // ignore
                continue;
            }

            in.unread(c);

            DataModel childModel;

            if (template == null) {
                // create child
                childModel = jsonFormatMap.createModel(in);

                if (childModel == null) {
                    throw new JsonParseException("cannot create child model");
                }
            } else {
                // create child from template
                childModel = template.duplicate(null, pairs);
            }

            IJsonFormat jsonFormat = jsonFormatMap.getJsonFormat(childModel);

            if (jsonFormat == null) {
                throw new JsonParseException("no json format for: " + childModel.getClass());
            }

            jsonFormat.parseJson(in, childModel);

            if (template == null && childModel instanceof DoubleModel) {
                DoubleModel dm = (DoubleModel) childModel;

                if (dm.getDouble() <= Long.MAX_VALUE && dm.getDouble() >= Long.MIN_VALUE && Math.abs(dm.getDouble()) == dm.getDouble()) {
                    childModel = new LongModel((long)dm.getDouble());
                }
            }

            array.append(childModel);            
        }
    }

    public void writeJson(DataModel source, OutputStream out) throws Exception {
        out.write('[');
        out.write('\n');

        DataModelEnumeration en = source.getEnumeration();
        boolean firstChild = true;

        while (en.hasMore()) {
            if (firstChild) {
                firstChild = false;
            } else {
                out.write(',');
            }

            DataModel childModel = en.next();

            IJsonFormat jsonFormat = jsonFormatMap.getJsonFormat(childModel);

            if (jsonFormat == null) {
                throw new Exception("no json format for: " + childModel.getClass());
            }

            jsonFormat.writeJson(childModel, out);

            out.write('\n');
        }

        out.write(']');
        out.write('\n');
    }

    public IJsonFormatMap getJsonFormatMap() {
        if (jsonFormatMap == null) {
            jsonFormatMap = DefaultJsonFormatMap.getInstance();
        }
        return jsonFormatMap;
    }
}