//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 *
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 *
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 *
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 *
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.dataviewer;

import java.awt.*;
import java.awt.event.*;
import java.util.Hashtable;

import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.*;

import com.nexirius.framework.FWLog;
import com.nexirius.framework.datamodel.*;

/**
 * This viewer is designed to show ArrayModel data. It is represented by a Swing JTable.
 *
 * @author Marcel Baumann
 */

public class ArrayViewer<T extends DataModel> extends DataViewer
{
    public static final String s_viewername = "ArrayViewer";
    public int editKey = KeyEvent.VK_ENTER; // FIX
    public ArrayViewerTableModel<T> tableModel = null;
    protected JTable table = null;
    protected MyListSelectionListener mySelectionListener;
    protected MySelector currentSelector;
    private HighlightListener highlightListener = new HighlightListener();
    protected ModelFlag selectedFlag = null;
    private boolean showPopupMenus = false;
    protected Hashtable<String, Boolean> sortOrder = new Hashtable<String, Boolean>();
    private SelectionHeaderRenderer selectionHeaderRenderer;
    protected TableColumn selectionColumn;
    private JPanel checkBoxCellRendererComponent;
    private JCheckBox cellRendererCheckBox;
    private HeaderCheckBox headerCheckBoxCellRendererComponent;

    protected boolean installSorter = true;

    /**
     * Creates an instance which is connected to the given model
     *
     * @param model The model which this viewer is showing.
     */
    public ArrayViewer(ArrayModel<T> model) {
        super(model);
    }

    public JComponent getCheckBoxCellRendererComponent() {
        if (checkBoxCellRendererComponent == null) {
            checkBoxCellRendererComponent = new JPanel();
            BoxLayout layout = new BoxLayout(checkBoxCellRendererComponent, BoxLayout.X_AXIS);
            checkBoxCellRendererComponent.setLayout(layout);
            cellRendererCheckBox = new JCheckBox("", false);
            cellRendererCheckBox.setOpaque(false);
            checkBoxCellRendererComponent.add(Box.createGlue());
            checkBoxCellRendererComponent.add(cellRendererCheckBox);
            checkBoxCellRendererComponent.add(Box.createGlue());
        }
        return checkBoxCellRendererComponent;
    }

    public void setSelectedFlag(ModelFlag selectedFlag)
    {
        this.selectedFlag = selectedFlag;
    }

    public ModelFlag getSelectedFlag()
    {
        return(selectedFlag);
    }

    public boolean isInstallSorter()
    {
        return installSorter;
    }

    public void setInstallSorter(boolean installSorter)
    {
        this.installSorter = installSorter;
    }

    public void setShowPopupMenus(boolean on) {
        showPopupMenus = on;
    }

    /**
     * Returns the associated data model
     */
    @SuppressWarnings("unchecked")
    public ArrayModel<T> getModel() {
        return (ArrayModel<T>) getDataModel();
    }

    /**
     * Creates and initialized the swing component
     */
    @Override
    public void create() {
        if (getFactory() == null) {
            throw new RuntimeException("Can't create a ArrayViewer with null factory (call setFactory())");
        }

        // First, create the TableModel and the JTable
        tableModel = createTableModel();
        createJTable();

        if (getLayout() == null) {
            table.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);
            table.setPreferredScrollableViewportSize(new java.awt.Dimension(300, 200));
        } else {
            table.setAutoResizeMode(JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS);
        }

        table.getSelectionModel().setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        JScrollPane scrollPane = new JScrollPane(table);
        scrollPane.setFocusable(false);
        scrollPane.getViewport().setFocusable(false);

        Component northComponent = createNorthComponent();
        Component southComponent = createSouthComponent();
        if (northComponent == null && southComponent == null)
        {
            // Der übliche Fall ==> Einfach the ScrollPane und die Tabelle verwenden
            setJComponent(scrollPane);
        }
        else
        {
            // Die Tabelle und die zusätzlichen Komponenten in ein Panel packen
            JPanel panel = new JPanel(new BorderLayout());
            panel.setFocusable(false);
            if (southComponent != null)
            {
                panel.add(southComponent, BorderLayout.SOUTH);
            }
            panel.add(scrollPane, BorderLayout.CENTER);
            if (northComponent != null)
            {
                panel.add(northComponent, BorderLayout.NORTH);
            }
            setJComponent(panel);
        }
        scrollPane.setMinimumSize(new Dimension(10, 10));

        table.addMouseListener(new MyMouseAdapter());

        if (installSorter)
        {
            table.getTableHeader().addMouseListener(createHeaderMouseAdapter());
        }
        table.addKeyListener(new MyKeyAdapter());
        mySelectionListener = new MyListSelectionListener();
        table.getSelectionModel().addListSelectionListener(mySelectionListener);

        updateSelection();

        getModel().addSoftDataModelListener(highlightListener);

        update();
        handleInvisibleFlag(null);

        if (getLayout() != null) {
            // set the column width
            getLayout().doLayout(this);
        }

        if (selectedFlag != null) {
            selectionColumn = table.getColumn(selectedFlag.getName());
            selectionColumn.setPreferredWidth(20);
            selectionColumn.setMaxWidth(20);
            selectionColumn.setCellRenderer(new TableCellRenderer() {
                @Override
                public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
                    boolean selected = false;
                    if (value instanceof Boolean) {
                        selected = ((Boolean) value).booleanValue();
                    }
                    boolean gray = false;

                    if (getModel().getSize() > row) {
                        gray = getModel().getDataModel(row).isGray();
                    }

                    JComponent ret = getCheckBoxCellRendererComponent();

                    cellRendererCheckBox.setSelected(selected);
                    cellRendererCheckBox.setEnabled(!gray);

                    if (isSelected) {
                        ret.setBackground(SystemColor.textHighlight);
                        ret.setForeground(SystemColor.textHighlightText);
                        ret.setOpaque(true);
                    } else {
                        ret.setForeground(SystemColor.text);
                        ret.setOpaque(false);
                    }

                    return ret;
                }
            });

            selectionColumn.setHeaderRenderer(new TableCellRenderer() {

                @Override
                public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
                    headerCheckBoxCellRendererComponent = getSelectionHeaderRenderer();
                    return (Component) headerCheckBoxCellRendererComponent;
                }
            });
        }
    }

    protected MouseAdapter createHeaderMouseAdapter() {
        return new MyHeaderMouseAdapter();
    }

    protected Component createNorthComponent()
    {
        return null;
    }

    protected Component createSouthComponent()
    {
        return null;
    }

    protected void createJTable()
    {
        table = new JTable(tableModel);
    }

    public HeaderCheckBox getSelectionHeaderRenderer() {
        if (selectionHeaderRenderer == null) {
            selectionHeaderRenderer = new SelectionHeaderRenderer();
        }

        return selectionHeaderRenderer;
    }

    public ArrayViewerTableModel<T> createTableModel() {
        if (getLayout() != null) {
            // create a table model from the layout information
            return new ArrayViewerTableModel<T>(this);
        }

        return new ArrayViewerTableModel<T>(this, getModel());
    }

    /**
     * When the associated data model changes its structure (number of columns) then
     * the table is reloaded.
     */
    @Override
    public synchronized void dataModelChangeStructure(DataModelEvent event) {
    }

    /**
     * The row of the subchild wich is the source of the value change event is updated.
     */
    @Override
    public synchronized void dataModelChangeValue(DataModelEvent event) {
        if (!isCreated()) {
            return;
        }

        handleInvisibleFlag(event);

        table.getSelectionModel().removeListSelectionListener(mySelectionListener);

        if (event.fromChild()) {
            int i = 0;
            DataModelVector<DataModel> path = event.getTransmitterVector();

            if (path.size() > 1) {
                // take the item just before the transmitter
                DataModel child = path.get(path.size() - 2);

                try {
                    i = getModel().getChildIndex(child);
                    tableModel.fireTableRowsUpdated(i, i);
                } catch (Exception ex) {
                    throw new RuntimeException(ex);
                }
            }
        } else {
            tableModel.fireTableDataChanged();
        }

        table.getSelectionModel().addListSelectionListener(mySelectionListener);

        updateSelection();
    }

    @Override
    public void update() {
    }

    /**
     * Returns the name of this viewer (for debugging)
     */
    @Override
    public String getViewerName() {
        return s_viewername;
    }

    /**
     * Call grabFocus() on the model which is associated to the current selection
     * @return The DataModel on which the Focus has been grabbed or null if there is no DataModel.
     */
    public DataModel grabFocusOnSelectedField() {
        DataModel rowModel = null;
        if (isCreated()) {
            int y = table.getSelectedRow();

            if (y != -1) {
                rowModel = getDataModelAtRow(y);

                if (rowModel != null) {
                    rowModel.grabFocus();
                }
            } else {
                FWLog.debug("Selected row index == -1");
            }
        }
        return rowModel;
    }

    public DataModel getDataModelAtRow(int y) {
        if (y < 0) {
            return null;
        }

        // y contains the row index as shown on the screen. This may not correspond 1:1 to the
        // index in the model. This is esp. true if sorting and/or filtering is involved!
        y = mapToModel(y);
        return getModel().getChildren().getItem(y);
    }

    /**
     * Maps a row index to the index into the model.
     * These two indizes may be different if the used table has sorting and filtering capabilities!
     * @param y
     * @return
     */
    protected int mapToModel(int y)
    {
        return y;
    }

    /**
     * Maps a model row index to the row index of the visible table.
     * These two indizes may be different if the used table has sorting and filtering capabilities!
     * @param y
     * @return
     */
    protected int mapToUserInterface(int y)
    {
        return y;
    }

    class MyKeyAdapter extends KeyAdapter {
        @Override
        public void keyPressed(KeyEvent e) {
            if (selectedFlag != null && e.getKeyCode() == KeyEvent.VK_SPACE) {
                DataModel highlightedItem = getModel().getHighlightedItem();
                if (highlightedItem != null && !highlightedItem.isGray()) {
                    highlightedItem.setFlag(selectedFlag, !highlightedItem.getFlag(selectedFlag));
                }
            }

            if (e.getKeyCode() == KeyEvent.VK_DELETE || e.getKeyCode() == KeyEvent.VK_BACK_SPACE) {
                DataModel highlightedItem = getModel().getHighlightedItem();

                if (highlightedItem != null) {
                    highlightedItem.fireEditEvent(highlightedItem, DataModelEvent.DELETE);
                }
            }

            if (e.getKeyCode() == editKey) {
                DataModel dataModel = grabFocusOnSelectedField();
                edit(dataModel);
            }

            // Ctrl+A
            if (selectedFlag != null && (e.getModifiers() & KeyEvent.CTRL_MASK) != 0 && e.getKeyCode() == KeyEvent.VK_A)
            {
                boolean on = (e.getModifiers() & KeyEvent.SHIFT_MASK) == 0;
                setAllSelected(on);
                return;
            }

            if ((e.getModifiers() & KeyEvent.ALT_MASK) == 0) {
                return;
            }

            if (e.getKeyCode() == KeyEvent.VK_UP) {
                ArrayModel<T> model = getModel();
                int index = model.getHighlightedItemIndex();

                if (model.getFlag(ModelFlag.REARRANGEABLE) && model.getSize() > 1 && index > 0) {
                    synchronized (model) {
                        T item = model.removeHighlightedItem();

                        model.insertItemAt(item, --index);
                        model.setHighlightedItem(index);
                        model.fireEditEvent(model, DataModelEvent.REARRANGED);
                    }
                }
            } else if (e.getKeyCode() == KeyEvent.VK_DOWN) {
                ArrayModel<T> model = getModel();
                int index = model.getHighlightedItemIndex();

                if (model.getFlag(ModelFlag.REARRANGEABLE) && index + 1 < model.getSize()) {
                    synchronized (model) {
                        T item = model.removeHighlightedItem();

                        model.insertItemAt(item, ++index);
                        model.setHighlightedItem(index);
                        model.fireEditEvent(model, DataModelEvent.REARRANGED);
                    }
                }
            }
        }

    }

    protected void edit(DataModel dataModel)
    {
        // Empty
    }

    public synchronized void select(int index) {
        if (isCreated()) {
            currentSelector = new MySelector(index);
            SwingUtilities.invokeLater(currentSelector);
        }
    }

    public void setAllSelected(boolean on) {
        ArrayModel<T> model = getModel();
        DataModelEnumeration<T> en = model.getEnumeration();

        while (en.hasMore()) {
            DataModel item = en.next();

            if (!item.isGray()) {
                item.setFlag(selectedFlag, on);
            }
        }
    }

    class MySelector implements Runnable {
        int index;

        MySelector(int i) {
            index = i;
        }

        @Override
        public synchronized void run() {
            synchronized (ArrayViewer.this) {
                if (currentSelector == this) {
                    table.getSelectionModel().removeListSelectionListener(mySelectionListener);

                    index = mapToUserInterface(index);
                    if (index < 0) {
                        table.getSelectionModel().clearSelection();
                    } else {
                        table.getSelectionModel().setAnchorSelectionIndex(index);
                        table.getSelectionModel().setLeadSelectionIndex(index);
                        table.getSelectionModel().setSelectionInterval(index, index);
                        table.scrollRectToVisible(table.getCellRect(index, 0, true));
                    }

                    table.getSelectionModel().addListSelectionListener(mySelectionListener);
                }
            }
        }
    }

    class MyMouseAdapter extends MouseAdapter {
        @Override
        public void mouseClicked(MouseEvent e) {
            if (e.getClickCount() == 2) {
                // Double click
                DataModel dataModel = grabFocusOnSelectedField();
                edit(dataModel);
            } else {
                // Single click
                int column = table.columnAtPoint(e.getPoint());
                int row = table.rowAtPoint(e.getPoint());

                if (row >= 0) {
                    DataModel model = (DataModel) table.getModel().getValueAt(row, -1);

                    if (model != null)
                    {
                        // Select the row
                        if (selectedFlag != null && column == 0 && !model.isGray()) {
                            model.setFlag(selectedFlag, !model.getFlag(selectedFlag));
                            //tableModel.fireTableCellUpdated(row, 0);
                        }
                    }
                }
            }
        }


        @Override
        public void mousePressed(MouseEvent e) {
            popupMenu(e);
        }

        @Override
        public void mouseReleased(MouseEvent e) {
            popupMenu(e);
        }

        private void popupMenu(MouseEvent e) {
            if (!showPopupMenus) {
                return;
            }
            if (e.isPopupTrigger()) {
                int row = table.rowAtPoint(e.getPoint());

                if (row >= 0) {
                    row = mapToModel(row);
                    getModel().setHighlightedItem(row);
                    DataModel model = getModel().getHighlightedItem();

                    if (model != null) {
                        DefaultTreePopupMenu menu = new DefaultTreePopupMenu();

                        // If anything is selected, then generate a menu
                        JPopupMenu popup = menu.createPopupMenu(factory, model);

                        if (popup != null) {
                            // Display popup menu at the given mouse position
                            popup.show(table, e.getX(), e.getY());
                        }
                    }
                }
            }
        }
    }

    class MyHeaderMouseAdapter extends MouseAdapter {
        @Override
        public void mouseClicked(MouseEvent e) {
            if (e.getClickCount() == 1) {

                Point point = e.getPoint();
                JTableHeader tableHeader = table.getTableHeader();
                int columnCount = table.getColumnCount();

                for (int i = 0; i < columnCount; ++i) {
                    Rectangle headerRect = tableHeader.getHeaderRect(i);

                    if (headerRect.contains(point)) {
                        String columnName = table.getColumnName(i);
                        TableColumn column = table.getColumn(columnName);

                        if (column == selectionColumn) {
//                            int mid = headerRect.x + headerRect.width / 2;

                            boolean val = headerCheckBoxCellRendererComponent.toggle();
                            //setAllSelected(point.x < mid);
                            setAllSelected(val);

                        }

                        int modelIndex = column.getModelIndex() - (selectedFlag == null ? 0 : 1);
                        String columnFieldName = tableModel.getColumnFieldName(modelIndex);

                        if (columnFieldName != null) {
                            Boolean normal = sortOrder.get(columnFieldName);
                            if (normal == null || normal.booleanValue()) {
                                getModel().sortByAttribute(columnFieldName);
                                sortOrder.put(columnFieldName, false);
                            } else {
                                getModel().sortByAttributeDesc(columnFieldName);
                                sortOrder.put(columnFieldName, true);
                            }
                        }
                    }
                }
            }
        }
    }



    class MyListSelectionListener implements ListSelectionListener {
        @Override
        public void valueChanged(ListSelectionEvent e) {
            if (table.getSelectionModel().getValueIsAdjusting()) {
                return;
            }

            updateSelectionIndexOfArrayModel();
        }
    }

    public void updateSelectionIndexOfArrayModel() {
        int index = table.getSelectionModel().getMinSelectionIndex();
        index = mapToModel(index);
        getModel().setHighlightedItem(index);
    }

    public synchronized void updateSelection() {
        int index = getModel().getHighlightedItemIndex();
        select(index);
    }

    class HighlightListener extends DataModelAdaptor {
        @Override
        public void dataModelEdit(DataModelEvent event) {
            if (event.getId() == DataModelEvent.HIGHLIGHT_CHANGED) {
                updateSelection();
            }
        }
    }

    public interface HeaderCheckBox {
        public boolean toggle();
    }

    class SelectionHeaderRenderer extends JPanel implements HeaderCheckBox {
        private static final long serialVersionUID = 7023637633531136344L;

        JCheckBox cb = new JCheckBox("", true);
        //JCheckBox off = new JCheckBox("", false);

        public SelectionHeaderRenderer() {
            setLayout(new GridLayout(1, 0));

            add(cb);
            //add(off);

//            addMouseListener(new MouseAdapter() {
//                public void mousePressed(MouseEvent e) {
//                    int mid = getSize().width;
//                    setAllSelected(e.getX() < mid);
//                }
//            });
        }

        @Override
        public void paint(Graphics g) {
            super.paint(g);
//            int width = getSize().width / 2;
//            int height = getSize().height;
//            Dimension d = new Dimension(width, height);
//            on.setSize(d);
//            on.setLocation(0, 0);
//            g.translate(0, 0);
//            on.paint(g);
//            on.setLocation(width, 0);
//            g.translate(width, 0);
//            off.paint(g);
        }

        @Override
        public boolean toggle() {
            boolean val = !cb.isSelected();
            cb.setSelected(val);

            return val;

        }
    }
}


