//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.dataviewer;

import com.nexirius.util.assertion.Assert;

/**
 * This class is exclusively as entries in the ViewerCreatorMap classes.
 *
 * @author Marcel Baumann
 * @see ViewerCreatorMap
 */
public class CreatorMapEntry {



    Class viewableClass;
    ViewerCreator viewerCreator = null;
    LayoutItem layoutItem = null;
    boolean isEditor = false;

    /**
     * Creates a new map entry.
     *
     * @param viewableClass The class of the data model or data model command which is associated with a creator
     * @param viewerCreator The creator instance which produces viewers or editors
     * @param layoutItem    The layout information which is used when creating a viewer or editor
     * @param isEditor      There are two kinds of map entries editors and viewers
     */
    public CreatorMapEntry(Class viewableClass, ViewerCreator viewerCreator, LayoutItem layoutItem, boolean isEditor) {
        this.viewableClass = viewableClass;
        this.viewerCreator = viewerCreator;
        this.layoutItem = layoutItem;
        this.isEditor = isEditor;
    }

    /**
     * Creates a new map entry.
     *
     * @param viewableClassName The class name of the data model or data model command which is associated with a creator
     * @param viewerCreator     The creator instance which produces viewers or editors
     * @param layoutItem        The layout information which is used when creating a viewer or editor
     * @param isEditor          There are two kinds of map entries editors and viewers
     */
    public CreatorMapEntry(String viewableClassName, ViewerCreator viewerCreator, LayoutItem layoutItem, boolean isEditor) {
        this(getClass(viewableClassName), viewerCreator, layoutItem, isEditor);
    }

    /**
     * Creates a new map entry.
     *
     * @param viewableClassName The class name of the data model or data model command which is associated with a creator
     * @param layoutItem        The layout information which is used when creating a viewer or editor
     * @param isEditor          There are two kinds of map entries editors and viewers
     */
    public CreatorMapEntry(String viewableClassName, LayoutItem layoutItem, boolean isEditor) {
        this(getClass(viewableClassName), null, layoutItem, isEditor);
    }

    /**
     * Creates a new map entry.
     *
     * @param viewableClassName The class name of the data model or data model command which is associated with a creator
     * @param viewerCreator     The creator instance which produces viewers or editors
     * @param layoutItem        The layout information which is used when creating a viewer or editor
     */
    public CreatorMapEntry(String viewableClassName, ViewerCreator viewerCreator, LayoutItem layoutItem) {
        this(viewableClassName, viewerCreator, layoutItem, viewerCreator.isEditor());
    }

    /**
     * Creates a new map entry.
     *
     * @param viewableClass The class of the data model or data model command which is associated with a creator
     * @param viewerCreator The creator instance which produces viewers or editors
     * @param layoutItem    The layout information which is used when creating a viewer or editor
     */
    public CreatorMapEntry(Class viewableClass, ViewerCreator viewerCreator, LayoutItem layoutItem) {
        this(viewableClass, viewerCreator, layoutItem, viewerCreator.isEditor());
    }

    /**
     * Creates a new map entry.
     *
     * @param viewableClass The class of the data model or data model command which is associated with a creator
     * @param viewerCreator The creator instance which produces viewers or editors
     */
    public CreatorMapEntry(Class viewableClass, ViewerCreator viewerCreator) {
        this(viewableClass, viewerCreator, null, viewerCreator.isEditor());
    }

    /**
     * Creates a new map entry.
     *
     * @param viewableClassName The class name of the data model or data model command which is associated with a creator
     * @param viewerCreator     The creator instance which produces viewers or editors
     */
    public CreatorMapEntry(String viewableClassName, ViewerCreator viewerCreator) {
        this(viewableClassName, viewerCreator, null);
    }

    /**
     * Creates a class from a class name
     */
    public static Class getClass(String viewableClassName) {
        try {
            return Class.forName(viewableClassName);
        } catch (Exception ex) {
            throw new RuntimeException("Can't load class '" + viewableClassName + "'");
        }
    }

    /**
     * Merge two map entries. This method is used if two map entries are used to introduce the layout and the viewer creator for the same
     * Viewable class type.
     */
    public void merge(CreatorMapEntry other) {
        Assert.pre(other.getViewableClass().equals(getViewableClass()), "Can't merge Creator map entry with different viewable classes: " + other.getViewableClass().getName() + " and " + getViewableClass().getName());
        Assert.pre(other.isEditor() == isEditor(), "Can't merge Creator map entry with different isEditor fields");

        if (other.getViewerCreator() != null) {
            setViewerCreator(other.getViewerCreator());
        }

        if (other.getLayoutItem() != null) {
            setLayoutItem(other.getLayoutItem());
        }
    }

    /**
     * Returns the viewables class name
     */
    public String getViewableClassName() {
        return viewableClass.getName();
    }

    /**
     * Returns the viewable class as Class instance
     */
    public Class getViewableClass() {
        return viewableClass;
    }

    /**
     * Returns the creator instance
     */
    public ViewerCreator getViewerCreator() {
        return this.viewerCreator;
    }

    /**
     * Returns the associated layout instance (or null)
     */
    public LayoutItem getLayoutItem() {
        return this.layoutItem;
    }

    /**
     * Set the viewer creator field
     */
    public void setViewerCreator(ViewerCreator viewerCreator) {
        this.viewerCreator = viewerCreator;
    }

    /**
     * Set the layout status field
     */
    public void setLayoutItem(LayoutItem layoutItem) {
        this.layoutItem = layoutItem;
    }

    /**
     * Returns true if this entry stands for editor creation
     */
    public boolean isEditor() {
        return isEditor;
    }
}
