//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.dataviewer;

import com.nexirius.framework.datamodel.*;

import javax.swing.*;
import java.awt.*;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

public class DataModelTreeComponent extends JPanel {
    BooleanModel folded = new BooleanModel(false, "");
    BooleanModel selected = new BooleanModel(false, "selected");
    JPanel headerPart;
    DataModelTreeHeaderPanel headerPanel;
    JPanel bodyPart;
    DataModel childModel;
    WeakReference treeViewer;
    HashMap treeComponentMap = new HashMap();
    boolean displayAsEditor = false;
    protected ViewerCreator viewerCreator;

    public DataModelTreeComponent(DataModelTreeViewer treeViewer, DataModel childModel) {
        this.treeViewer = new WeakReference(treeViewer);
        this.childModel = childModel;

        folded.addDataModelListener(new FoldedListener());
        selected.addDataModelListener(new SelectionListener());
    }

    public void init() {
        headerPanel = createHeaderPanel();
        headerPart = createHeaderPart();
        if (!folded.getBoolean()) {
            bodyPart = createBodyPart();
        }

        layoutAll();
    }

    public BooleanModel getFolded() {
        return folded;
    }

    public BooleanModel getSelected() {
        return selected;
    }

    public JPanel getHeaderPart() {
        return headerPart;
    }

    public DataModelTreeHeaderPanel getHeaderPanel() {
        return headerPanel;
    }

    public void layoutAll() {
        removeAll();
        setLayout(new BorderLayout());
        add(getHeaderPart(), BorderLayout.NORTH);

        if (bodyPart != null) {
            add(bodyPart, BorderLayout.CENTER);
        }
    }

    public void setDisplayAsEditor(boolean displayAsEditor, ViewerCreator viewerCreator) {
        if (this.displayAsEditor != displayAsEditor || this.viewerCreator != viewerCreator) {
            this.displayAsEditor = displayAsEditor;
            this.viewerCreator = viewerCreator;
            bodyPart = createBodyPart();
            layoutAll();
            treeComponentMap.clear();
            getChildModel().fireValueChange();
        }
    }

    public void setDisplayAsEditor(boolean displayAsEditor) {
        setDisplayAsEditor(displayAsEditor, null);
    }

    public boolean getDisplayAsEditor() {
        return displayAsEditor;
    }

    public JPanel createHeaderPart() {
        JPanel ret = new JPanel();
        ret.setLayout(new GridLayout(1, 0));

        ret.add(getHeaderPanel());
        ret.setOpaque(true);

        return ret;
    }

    public DataModelTreeHeaderPanel createHeaderPanel() {
        return new DataModelTreeHeaderPanel(this);
    }

    public JPanel createBodyPart() {
        if (displayAsEditor || !(childModel instanceof DataModelContainer)) {
            return createEditorBodyPart();
        } else {
            JPanel ret = new JPanel();
            ret.setLayout(null);
            ret.setOpaque(true);

            return ret;
        }
    }

    public JPanel createEditorBodyPart() {
        JPanel ret = new JPanel();
        ret.setLayout(null);
        ret.setOpaque(true);

        ret.setLayout(new GridLayout(1, 1));

        try {
            if (viewerCreator == null) {
                ret.add(getTreeViewer().getFactory().createDefaultEditor(childModel).getJComponent());
            } else {
                ret.add(getTreeViewer().getFactory().createViewer(viewerCreator, childModel).getJComponent());
            }

            return ret;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public void addTreeComponent(DataModelTreeComponent tc) {
        treeComponentMap.put(tc.getModelKey(), tc);
    }

    public DataModelTreeComponent getTreeComponentFor(DataModel dm) {
        return (DataModelTreeComponent) treeComponentMap.get(new ModelKey(dm));
    }

    private ModelKey getModelKey() {
        return new ModelKey(childModel);
    }

    /**
     * @return > 0 if height has changed
     */
    public int updateComponent(DataModel source) {
        // go thru all children and updateComponent

        if (displayAsEditor || childModel.getChildren() == null) {
            headerPanel.setHasChildren(displayAsEditor ? true : false);
            updateHeader();

            return 0;
        }

        int changeCount = 0;
        DataModelEnumeration enumVar = getChildrenEnumeration();
        boolean treeChildrenChanged = false;
        ArrayList childComponentList = new ArrayList();

        while (enumVar.hasMore()) {
            DataModel child = enumVar.next();
            DataModelTreeComponent childComponent = getTreeComponentFor(child);

            if (childComponent == null) {
                childComponent = getTreeViewer().createDataModelTreeViewer(child);
                addTreeComponent(childComponent);
                treeChildrenChanged = true;
                ++changeCount;
            } else if (child == source) {
                childComponent.updateHeader();
            }

            changeCount += childComponent.updateComponent(source);

            childComponentList.add(childComponent);
        }

        if (treeComponentMap.size() > childComponentList.size()) {
            treeChildrenChanged = true;
            // remove all unused components from hashmap
            treeComponentMap.clear();

            Iterator iter = childComponentList.iterator();

            while (iter.hasNext()) {
                addTreeComponent((DataModelTreeComponent) iter.next());
            }
        }

        if (treeChildrenChanged) {
            if (bodyPart != null) {
                bodyPart.removeAll();
            }

            Iterator it = childComponentList.iterator();

            while (it.hasNext()) {
                DataModelTreeComponent treeComponent = (DataModelTreeComponent) it.next();

                if (bodyPart != null) {
                    bodyPart.add(treeComponent);
                }
            }
        }

        headerPanel.setHasChildren(childComponentList.size() > 0);

        relayout();

        return changeCount;
    }

    public void updateHeader() {
        headerPanel.update();
    }

    private DataModelEnumeration getChildrenEnumeration() {
        return new ChildrenDataModelEnumeration(childModel);
    }

    static class ChildrenDataModelEnumeration extends DataModelEnumeration {
        ChildrenDataModelEnumeration subEnum = null;
        DataModelEnumeration enumVar;
        DataModel nextInstance;

        public ChildrenDataModelEnumeration(DataModel child) {
            this.enumVar = child.getEnumeration();
            nextInstance = nextInstance();
        }

        public boolean hasMore() {
            return nextInstance != null;
        }

        public DataModel next() {
            DataModel ret = nextInstance;

            nextInstance = nextInstance();

            return ret;
        }

        public DataModel nextInstance() {
            if (subEnum != null && subEnum.hasMore()) {
                return subEnum.next();
            } else {
                subEnum = null;
            }

            if (enumVar != null && enumVar.hasMore()) {
                DataModel ret = enumVar.next();

                while (ret.getFlag(ModelFlag.INVISIBLE_TREE_NODE)) {
                    if (enumVar.hasMore()) {
                        ret = enumVar.next();
                    } else {
                        return null;
                    }
                }

                if (ret.getFlag(ModelFlag.SKIP_TREE_NODE)) {
                    subEnum = new ChildrenDataModelEnumeration(ret);

                    return nextInstance();
                }


                if (!enumVar.hasMore()) {
                    enumVar = null;
                }

                return ret;
            }

            return null;
        }
    }

    private void relayout() {
        if (bodyPart == null) {
            return;
        }

        bodyPart.setLayout(null);
        int y = getTreeViewer().getYGap();
        int maxWidth = 0;
        Component[] components = bodyPart.getComponents();

        for (int i = 0; i < components.length; i++) {
            if (components[i] instanceof DataModelTreeComponent) {
                DataModelTreeComponent treeComponent = (DataModelTreeComponent) components[i];

                treeComponent.setSize(treeComponent.getPreferredSize());
                Rectangle bounds = treeComponent.getBounds();

                bounds.x = getTreeViewer().getXGap();
                bounds.y = y;
                treeComponent.setBounds(bounds);

                y += bounds.height + getTreeViewer().getYGap();
                maxWidth = Math.max(maxWidth, bounds.width);
            }
        }

        Dimension newSize = new Dimension(maxWidth + getTreeViewer().getXGap(), y);
        bodyPart.setSize(newSize);
        bodyPart.setPreferredSize(newSize);
        this.revalidate();
    }

    public void updateViewer() {
        getTreeViewer().update();
    }

    public DataModelTreeViewer getTreeViewer() {
        return (DataModelTreeViewer) treeViewer.get();
    }

    public void setSelected(boolean on) {
        selected.setBoolean(on);
    }

    public void toggleSelected(boolean isControlDown, boolean isShiftDown) {
        if (!isShiftDown && !isControlDown) {
            getTreeViewer().getRootTreeComponent().setSelectedRecursive(false);
        }

        selected.setBoolean(!selected.getBoolean());
        getTreeViewer().callDataModelTreeViewerListenerSelection(this);
    }

    public void setFolded(boolean on) {
        folded.setBoolean(on);
    }

    /**
     *
     * @param isControlDown if true toggles recursively
     * @param isShiftDown if true toggles recursively
     */
    public void toggleFolded(boolean isControlDown, boolean isShiftDown) {
        boolean on = !folded.getBoolean();

        if (isShiftDown || isControlDown) {
            setFoldedRecursive(on);
            getTreeViewer().callDataModelTreeViewerListenerFolding(this, true);
        } else {
            folded.setBoolean(on);
            getTreeViewer().callDataModelTreeViewerListenerFolding(this, false);
        }
    }

    public void setSelectedRecursive(boolean on) {
        setSelected(on);

        DataModelEnumeration enumVar = getChildrenEnumeration();

        while (enumVar.hasMore()) {
            DataModel child = enumVar.next();
            DataModelTreeComponent treeComponent = getTreeComponentFor(child);

            if (treeComponent != null) {
                treeComponent.setSelectedRecursive(on);
            }
        }
    }

    public void setFoldedRecursive(boolean on) {
        setFolded(on);

        DataModelEnumeration enumVar = getChildrenEnumeration();

        while (enumVar.hasMore()) {
            DataModel child = enumVar.next();
            DataModelTreeComponent treeComponent = getTreeComponentFor(child);

            if (treeComponent != null) {
                treeComponent.setFoldedRecursive(on);
            }
        }
    }

    public DataModel getChildModel() {
        return childModel;
    }

    public void getSelectedDataModels(DataModelVector v) {
        if (selected.getBoolean()) {
            v.append(getChildModel());
        }

        DataModelEnumeration enumVar = getChildrenEnumeration();

        if (childModel.getChildren() == null) {
            return;
        }

        while (enumVar.hasMore()) {
            DataModel child = enumVar.next();
            DataModelTreeComponent childComponent = getTreeComponentFor(child);

            if (childComponent != null) {
                childComponent.getSelectedDataModels(v);
            }
        }
    }

    public void getFoldedDataModels(DataModelVector v, boolean isFolded) {
        if (folded.getBoolean() == isFolded) {
            v.append(getChildModel());
        }

        DataModelEnumeration enumVar = getChildrenEnumeration();

        if (childModel.getChildren() == null) {
            return;
        }

        while (enumVar.hasMore()) {
            DataModel child = enumVar.next();
            DataModelTreeComponent childComponent = getTreeComponentFor(child);

            if (childComponent != null) {
                childComponent.getFoldedDataModels(v, isFolded);
            }
        }
    }

    public boolean isFolded() {
        return folded.getBoolean();
    }

    class FoldedListener extends DataModelAdaptor {
        public void dataModelChangeValue(DataModelEvent event) {
            if (!folded.getBoolean() && bodyPart == null) {
                bodyPart = createBodyPart();
                layoutAll();
            }
            headerPanel.setFolded(folded.getBoolean());
            bodyPart.setVisible(!folded.getBoolean());
            updateViewer();
        }
    }

    class SelectionListener extends DataModelAdaptor {
        public void dataModelChangeValue(DataModelEvent event) {
            headerPanel.setSelected(selected.getBoolean());
        }
    }

    public static class ModelKey {
        Object model;

        public ModelKey(DataModel model) {
            this.model = model;
        }

        public boolean equals(Object obj) {
            return model == ((ModelKey) obj).model;
        }

        public int hashCode() {
            return model.hashCode();
        }
    }

//    public class BodyPanel extends JPanel {
//        public void paint(Graphics g) {
//            super.paint(g);
//            Component[] components = getComponents();
//
//            int y = 0;
//            int oldy = 0;
//            int xbase = 7;
//            int ybase = 13;
//
//            for (int i = 0; i < components.length; i++) {
//                Component treeComponent = components[i];
//                Rectangle bounds = treeComponent.getBounds();
//
//                g.drawLine(xbase, oldy, xbase, y + ybase);
//                oldy = y + ybase;
//                g.drawLine(xbase, y + ybase, bounds.x, y + ybase);
//
//                y += bounds.height + getTreeViewer().getYGap();
//            }
//        }
//    }
}
