//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 *
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 *
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 *
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 *
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.dataviewer;

import java.util.ConcurrentModificationException;
import java.util.Vector;

import com.nexirius.framework.datamodel.*;

/**
 * This viewer is designed to show ArrayModel data. Only data models which have the specified filteredFlag set are displayed.
 *
 * @author Marcel Baumann
 */

public class FilteredArrayViewer<T extends DataModel> extends ArrayViewer<T>
{
    ModelFlag filteredFlag;
    protected FilteredArrayTableModel filteredArrayTableModel;

    /**
     * Creates an instance which is connected to the given model
     *
     * @param model The model which this viewer is showing.
     */
    public FilteredArrayViewer(ArrayModel<T> model, ModelFlag filteredFlag) {
        super(model);
        this.filteredFlag = filteredFlag;
    }

    @Override
    public ArrayViewerTableModel createTableModel() {
        if (getLayout() != null) {
            // create a table model from the layout information
            filteredArrayTableModel = new FilteredArrayTableModel(this);
        } else {
            filteredArrayTableModel = new FilteredArrayTableModel(this, getModel());
        }

        return filteredArrayTableModel;
    }

    @Override
    public DataModel getDataModelAtRow(int y) {
        if (y < 0) {
            return null;
        }

        return getModel().getChildren().getItem(filteredArrayTableModel.getFilteredRowIndex(y));
    }

    @Override
    public synchronized void dataModelChangeValue(DataModelEvent event) {
        super.dataModelChangeValue(event);
        if (filteredArrayTableModel != null) {
            filteredArrayTableModel.setDoRecalculate(true);
        }
    }

    @Override
    public synchronized void updateSelection() {
        int index = getModel().getHighlightedItemIndex();

        if (index < 0) {
            select(-1);
        } else {
            select(filteredArrayTableModel.getRealRowIndex(index));
        }
    }

    @Override
    public void updateSelectionIndexOfArrayModel() {
        int selectionIndex = table.getSelectionModel().getMinSelectionIndex();

        if (selectionIndex < 0) {
            getModel().setHighlightedItem(-1);
        } else {
            getModel().setHighlightedItem(filteredArrayTableModel.getFilteredRowIndex(selectionIndex));
        }
    }

    public class FilteredArrayTableModel extends ArrayViewerTableModel {
        private static final long serialVersionUID = -3225171386877617873L;

        boolean doRecalculate = true;
        Vector<Integer> filtered = new Vector<Integer>();

        public FilteredArrayTableModel(ArrayViewer<T> arrayViewer,  ArrayModel<T> model) {
            super(arrayViewer, model);
            recalculate();
        }

        public FilteredArrayTableModel(ArrayViewer<T> arrayViewer)
        {
            super(arrayViewer);
            recalculate();
        }

        public void setDoRecalculate(boolean doRecalculate) {
            this.doRecalculate = doRecalculate;
        }

        private void recalculate() {
            if (!doRecalculate) {
                return;
            }
            int exceptionCount = 0;
            ConcurrentModificationException exception = null;

            while (exceptionCount < 10) {
                exception = null;
                try {
                    filtered.clear();
                    DataModelEnumeration<? extends DataModel> enumVar = getDataModel().getEnumeration();
                    int index = 0;

                    while (enumVar.hasMore()) {
                        DataModel model = enumVar.next();
                        if (model.getFlag(filteredFlag)) {
                            filtered.add(index);
                        }
                        index++;
                    }

                    doRecalculate = false;
                    break;
                } catch (ConcurrentModificationException ex) {
                    exception = ex;
                    ++exceptionCount;
                }
            }

            if (exception != null) {
                throw exception;
            }
        }

        private int getFilteredRowIndex(int y) {
            recalculate();
            int ret =  filtered.get(y);
            return ret;
        }

        @Override
        public Object getValueAt(int y, int x) {
            recalculate();
            return super.getValueAt(getFilteredRowIndex(y), x);
        }

        @Override
        public int getRowCount() {
            recalculate();
            return filtered.size();
        }

        public int getRealRowIndex(int selectionIndex) {
            recalculate();

            int ret = filtered.indexOf(selectionIndex);

            return ret;
        }
    }
}

