//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.htmlview;

import com.nexirius.framework.FWLog;
import com.nexirius.framework.datamodel.DataModel;
import com.nexirius.framework.datamodel.DataModelVector;
import com.nexirius.framework.datamodel.BooleanModel;
import com.nexirius.framework.htmlview.function.HTMLFunction;
import com.nexirius.util.StringVector;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;

public class HTMLSessionVariable {
    public static final String BOOLEAN_PARAMETER_PREFIX = "boolean__";

    private String state;
    private String event;
    private String button;
    private String child;
    private boolean duplicate;
    private String template;
    private Boolean isEditor = null;
    private Integer index = null;
    private HTMLStateMachine stateMachine;
    private HTMLResolver resolver;
    private Vector backwardStateVector;
    private StringVector selectedChildren;
    private int maxNumberOfBackwardStates;
    private HttpServletRequest request;
    protected HttpServletResponse response;
    private RequestParameterMap requestParameterMap = new RequestParameterMap();

    public HTMLSessionVariable(int maxNumberOfBackwardStates, HTMLStateMachine stateMachine, HTMLResolver resolver) {
        this.maxNumberOfBackwardStates = maxNumberOfBackwardStates;
        this.stateMachine = stateMachine;
        this.resolver = resolver;
        backwardStateVector = new Vector();
    }

    public void setIsEditor(String isEditorString) {
        isEditor = null;
        if (isEditorString != null) {
            if (isEditorString.equals(HTMLParser.EDIT)) {
                isEditor = Boolean.TRUE;
            } else if (isEditorString.equals(HTMLParser.VIEW)) {
                isEditor = Boolean.FALSE;
            }
        }
    }

    public void setIndex(String indexString) {
        index = null;
        if (indexString != null) {
            try {
                index = new Integer(indexString);
            } catch (NumberFormatException ex) {
                index = null;
            }
        }
    }

    public HTMLState getActState() {
        return (HTMLState) stateMachine.getActState();
    }

    /**
     * Access the state, which is displayed if the actual command has finished execution and returned true
     */
    public HTMLState getNextState() {
        return (HTMLState) getStateMachine().getActState().getNextStateForEvent(getEvent());
    }

    /**
     * Access the next state (if the actual command has finished execution and returned true) and return its associated DataModel
     *
     * @return the DataModel, which will be displayed with the next state
     */
    public DataModel getModelOfNextState() throws Exception {
        if (child != null) {
            if (getActModel() != null) {
                return getActModel().getChild(child);
            }
        }
        return getNextState().getModel();
    }

    /**
     * Access the next state (if the actual command has finished execution and returned true) and set its associated DataModel
     *
     * @param model The DataModel, which will be displayed with the next state
     */
    public void setModelOfNextState(DataModel model) {
        getNextState().setModel(model);
    }

    public DataModel getActModel() {
        return getActState().getModel();
    }

    /**
     * This is the main model which held in the state machine and which need not be attached to any state in the state machine
     *
     * @return The actual main model
     */
    public DataModel getApplicationModel() {
        return stateMachine.getApplicationModel();
    }

    /**
     * clear the StringVector which holds the names of the currently selelcted children of the actual data model
     */
    public void clearSelectedChildren() {
        selectedChildren = null;
    }

    /**
     * This method is called, when using the $!select() commmand to register a child which was selected by ticking the
     * corresponding check box.
     *
     * @param parameter request parameter
     */
    public void addSelectedChild(String parameter) {
        if (selectedChildren == null) {
            selectedChildren = new StringVector();
        }

        String childName = requestParameterMap.translateParam(parameter);

        FWLog.debug("add " + childName);
        selectedChildren.append(childName);
    }

    /**
     * Access all children of the current data model as DataModelVector which have been selelcted by ticking the
     * checkboxes which have been created with the $!select() command
     */
    public DataModelVector getSelectedChildren() {
        DataModelVector ret = new DataModelVector();
        DataModel parentModel = getActModel();

        FWLog.debug("selectedChildren " + selectedChildren);
        FWLog.debug("parentModel " + parentModel.getFieldName());

        if (selectedChildren == null) {
            return ret;
        }

        for (String childName = selectedChildren.firstItem(); childName != null; childName = selectedChildren.nextItem())
        {
            try {
                ret.append(parentModel.getChild(childName));
            } catch (Exception e) {
                //ignore
                throw new RuntimeException(e);
            }
        }

        return ret;
    }

    public String getState() {
        return state;
    }

    public String getEvent() {
        return event;
    }

    public String getButton() {
        return button;
    }

    public String getChild() {
        return child;
    }

    public boolean getDuplicate() {
        return duplicate;
    }

    public String getTemplate() {
        return template;
    }

    public Boolean isEditor() {
        return isEditor;
    }

    public Integer getIndex() {
        return index;
    }

    public HTMLResolver getResolver() {
        return resolver;
    }

    public HTMLStateMachine getStateMachine() {
        return stateMachine;
    }

    public void setEvent(String event) {
        this.event = event;
    }

    public void init(HttpServletRequest request, HttpServletResponse response) {
        this.request = request;
        this.response = response;
        state = request.getParameter(HTMLFunction.PARAMETER_STATE);
        event = request.getParameter(HTMLFunction.PARAMETER_EVENT);
        child = request.getParameter(HTMLFunction.PARAMETER_CHILD);
        duplicate = "true".equalsIgnoreCase(request.getParameter(HTMLFunction.PARAMETER_DUPLICATE));
        button = null;
        template = request.getParameter(HTMLFunction.PARAMETER_TEMPLATE);
        setIsEditor(request.getParameter(HTMLFunction.PARAMETER_ISEDITOR));
        setIndex(request.getParameter(HTMLFunction.PARAMETER_INDEX));

        Enumeration e = request.getParameterNames();

        while (e.hasMoreElements()) {
            String p = (String) e.nextElement();

            FWLog.debug("PARAMETER: " + p + "=" + request.getParameter(p));

            if (p.startsWith(HTMLFunction.PARAMETER_BUTTON)) {
                button = request.getParameter(p);
                int lastIndexOfAt = p.lastIndexOf('@');
                event = p.substring(p.indexOf('@') + 1, lastIndexOfAt);
                state = p.substring(lastIndexOfAt + 1);
            }

            if (p.startsWith(HTMLFunction.PARAMETER_SELECTED)) {
                addSelectedChild(p.substring(HTMLFunction.PARAMETER_SELECTED.length()));
            }
        }

        FWLog.debug("state = " + state);
        FWLog.debug("event = " + event);
        FWLog.debug("button = " + button);
        FWLog.debug("template = " + template);
        FWLog.debug("child = " + child);
        FWLog.debug("isEditor = " + isEditor);
        FWLog.debug("index = " + index);
    }

    /**
     * reset the state machine to the beginning state
     */
    public void resetStateMachine() {
        stateMachine.reset();
        backwardStateVector.removeAllElements();
        storeActState();
    }

    /**
     * This method performs the next step in the actual state machine
     */
    public void doStateTransition() {
        if (getEvent() != null) {
            try {
                getStateMachine().step(getEvent());
            } catch (Exception ex) {
                // ignore
                FWLog.debug("step denied = " + ex);
            }
        }
    }

    /**
     * Stores the actual state into the backward buffer.
     */
    public void storeActState() {
        backwardStateVector.insertElementAt(stateMachine.getActState(), 0);

        while (backwardStateVector.size() > maxNumberOfBackwardStates) {
            backwardStateVector.remove(backwardStateVector.size() - 1);
        }
    }

    /**
     * Go back in the backward state buffer by one state.
     *
     * @return actual state
     */
    public HTMLState goBackOneState() throws Exception {
        if (backwardStateVector.size() > 1) {
            backwardStateVector.remove(0);
            HTMLState jumpState = (HTMLState) backwardStateVector.get(0);
            backwardStateVector.remove(0);
            stateMachine.setActState(jumpState);
        }

        return getActState();
    }

    /**
     * Go back in the backward state buffer and find the actual state. If the state is found then the
     * state machine performs a jump to that state. If the state is not found, then the state machine
     * performs a reset.
     *
     * @return
     * @throws Exception
     */
    public HTMLState goBackwardState(String targetStateName)
            throws Exception {
        Enumeration e = backwardStateVector.elements();
        HTMLState jumpState = null;
        int i = 0;

        if (targetStateName != null) {
            while (e.hasMoreElements()) {
                HTMLState s = (HTMLState) e.nextElement();

                if (s.getName().equals(targetStateName)) {

                    jumpState = s;
                    stateMachine.setActState(jumpState);
                    break;
                }
                ++i;
            }
        }

        if (jumpState == null) {
            // no such state in backward state buffer
            resetStateMachine();
            jumpState = (HTMLState) stateMachine.getActState();
        } else {
            while (i > 0) {
                backwardStateVector.remove(0);
                --i;
            }
        }

        return jumpState;
    }

    public void setChildName(String childName) {
        child = childName;
    }

    public HttpServletRequest getRequest() {
        return request;
    }

    public HttpServletResponse getResponse() {
        return response;
    }

    public String getRequestParameter(String parameterName) {
        return getRequest().getParameter(parameterName);
    }

    public void mapRequestParametersToModel(HttpServletRequest request) {
        DataModel actModel = getActState().getModel();

        if (actModel != null) {
            mapRequestParametersToModel(request, actModel);
        }

    }

    /**
     * Map the request parameters to the actual DataModel which is held in the current state (sessionVariable.getActState().getModel()).
     * In general the name and values of the parameters are used like this: actModel.setChildText(parameterName, value);
     *
     * @param request
     * @param actModel
     */
    public void mapRequestParametersToModel(HttpServletRequest request, DataModel actModel) {
        Enumeration enumVar = request.getParameterNames();
        String falseParameter = null;

        while (enumVar.hasMoreElements()) {
            String parameterName = (String) enumVar.nextElement();
            String value = request.getParameter(parameterName);
            String fieldName = requestParameterMap.translateParam(parameterName);

            if (falseParameter != null && !falseParameter.equals(parameterName)) {
                try {
                    ((BooleanModel) actModel.getChild(fieldName)).setBoolean(false);
                    falseParameter = null;
                } catch (Throwable ex) {
                    // ignore
                }
            }

            if (parameterName.startsWith(HTMLFunction.PARAMETER_BUTTON)) {

                continue;
            } else if (parameterName.equals("this")) {
                parameterName = null;
            } else if (parameterName.startsWith(BOOLEAN_PARAMETER_PREFIX)) {
                falseParameter = parameterName.substring(BOOLEAN_PARAMETER_PREFIX.length());
                continue;
            }

            // check whether child is a BooleanModel
            try {
                DataModel child = actModel.getChild(fieldName);

                if (child instanceof BooleanModel) {
                    ((BooleanModel) child).setBoolean(true);
                } else {
                    actModel.setChildText(fieldName, value);
                }
            } catch (Exception ex) {
                // ignore
            }

            FWLog.debug("Setting child value " + parameterName + "=" + value);
        }
    }

    public String getRequestParameterForFieldName(String fieldName) {
        return requestParameterMap.translateField(fieldName);
    }

    class RequestParameterMap {
        int counter = 0;
        private Hashtable param2field = new Hashtable();
        private Hashtable field2param = new Hashtable();
        private static final String PARAMETER_PREFIX = "p__";

        public String translateField(String field) {
            String param = (String) field2param.get(field);

            if (param == null) {
                ++counter;
                param = PARAMETER_PREFIX + counter;
                field2param.put(field, param);
                param2field.put(param, field);
            }

            return param;
        }

        public String translateParam(String param) {
            if (!param.startsWith(PARAMETER_PREFIX)) {
                return param;
            }

            String fieldName = (String) param2field.get(param);

            if (fieldName == null) {
                throw new RuntimeException("cannot translate parameter:" + param);
            }

            return fieldName;
        }
    }
}
