//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.jdbc;

import com.nexirius.framework.datamodel.DataModel;

import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Date;

/**
 * Singleton implementation for a database connection handler. Maps type names to DatabaseTableMapping instances.
 */
public class JdbcConnectionHandler {
    private static JdbcConnectionHandler jdbcConnectionHandler;

    private DataSource dataSource;
    private String url;
    private String username;
    private String password;
    private HashMap databaseTableMap = new HashMap();
    private Connection updateConnection;
    private Connection readConnection;
    private Date lastAccessToConnection = new Date();
    private static long primaryKey;
    public static int TIMEOUT_SECONDS = -1; // never time out

    public JdbcConnectionHandler(DataSource dataSource, String username, String password) {
        this.dataSource = dataSource;
        this.username = username;
        this.password = password;
    }

    public JdbcConnectionHandler(String url, String username, String password) {
        this.url = url;
        this.username = username;
        this.password = password;
    }

    public static void init(JdbcConnectionHandler handler) {
        jdbcConnectionHandler = handler;
    }

    public static JdbcConnectionHandler instance() {
        return jdbcConnectionHandler;
    }

    public synchronized Connection getConnection(boolean readOnly) throws SQLException {
        if (TIMEOUT_SECONDS >= 0 && (readConnection != null || updateConnection != null)) {
            // check for timed out

            Date now = new Date();
            if (now.getTime() - TIMEOUT_SECONDS * 1000 > lastAccessToConnection.getTime()) {
                closeConnectionsLater(readConnection, updateConnection);
                readConnection = null;
                updateConnection = null;
                lastAccessToConnection = now;
            }
        }

        if (readOnly) {
            if (readConnection == null || readConnection.isClosed()) {
                readConnection = getConnection();
                readConnection.setReadOnly(true);
            }

            return readConnection;
        }

        if (updateConnection == null || updateConnection.isClosed()) {
            updateConnection = getConnection();
        }

        return updateConnection;
    }

    private Connection getConnection() throws SQLException {
        if (dataSource == null) {
            return username == null ? DriverManager.getConnection(url) : DriverManager.getConnection(url, username, password);
        }

        return username == null ? dataSource.getConnection() : dataSource.getConnection(username, password);
    }

    public synchronized void closeConnection() throws SQLException {
        if (updateConnection != null) {
            updateConnection.close();
            updateConnection = null;
        }
        if (readConnection != null) {
            readConnection.close();
            readConnection = null;
        }
    }

    public synchronized void closeConnectionsLater(final Connection connection1, final Connection connection2) {
        new Thread(new Runnable() {
            public void run() {
                try {
                    Thread.sleep(30000);
                } catch (InterruptedException e) {
                    throw new RuntimeException(e);  //To change body of catch statement use File | Settings | File Templates.
                }
                if (connection1 != null) {
                    try {
                        connection1.close();
                    } catch (SQLException e) {
                        throw new RuntimeException(e);  //To change body of catch statement use File | Settings | File Templates.
                    }
                }
                if (connection2 != null) {
                    try {
                        connection2.close();
                    } catch (SQLException e) {
                        throw new RuntimeException(e);  //To change body of catch statement use File | Settings | File Templates.
                    }
                }
            }
        }, "closeConnectionsLater").start();
    }

    /**
     * Register standard mapping (see DatabaseTableMapping.createStandardMapping())
     *
     * @param dataModel
     * @return the newly created mapping
     */
    public DatabaseTableMapping registerDatabaseTableMapping(DataModel dataModel) {
        DatabaseTableMapping standardMapping = DatabaseTableMapping.createStandardMapping(dataModel);

        registerDatabaseTableMapping(dataModel.getClass().getName(), standardMapping);

        return standardMapping;
    }

    /**
     * Register simple DataModel to database table mapping.<br>
     * <pre>
     * EXAMPLE
     * <p/>
     * DataModel:
     * class MyDataModel {
     *  public MyDataModel() {
     *    append(new StringModel(0, "name"));
     *    append(new IntModel(0, "value"));
     *  }
     * }
     * <p/>
     * DB Table Name: NAME_VALUE_TABLE
     * DB Fields:
     * LONG F_UID
     * VARCHAR(30) F_NAME
     * INTEGER F_VALUE
     * <p/>
     * call:
     * registerDatabaseTableMapping(MyDataModel.class.getName(), "NAME_VALUE_TABLE", "F_UID", new String[]{"name","value"}, new String[]{"F_NAME","F_VALUE"});
     * </pre>
     *
     * @param dataModelTypeName       the name of the DataModel class
     * @param dbTableName             the DB table name
     * @param primaryKeyFieldName     the name of the primary key (the value of the primary key is held in the instance name of the data model getInstanceName())
     * @param dataModelAttributeNames holds a list of data model child names (also nested names are supported) which should be stored in database fields
     * @param dbFieldNames            the database field names associated to the data model attribute names
     * @return the newly created database table mapping
     */
    public DatabaseTableMapping registerDatabaseTableMapping(String dataModelTypeName, String dbTableName, String primaryKeyFieldName, String[] dataModelAttributeNames, String[] dbFieldNames) {
        DatabaseTableMapping mapping = new DatabaseTableMapping(dbTableName, primaryKeyFieldName, dataModelAttributeNames, dbFieldNames);

        registerDatabaseTableMapping(dataModelTypeName, mapping);

        return mapping;
    }

    public void registerDatabaseTableMapping(String dataModelTypeName, DatabaseTableMapping mapping) {
        mapping.setJdbcConnectionHandler(this);
        mapping.setDataModelClassName(dataModelTypeName);
        databaseTableMap.put(dataModelTypeName, mapping);
    }

    public DatabaseTableMapping getDatabaseTableMapping(String typeName) {
        //TODO check if OK set connection when returning
        DatabaseTableMapping dtm = (DatabaseTableMapping)databaseTableMap.get(typeName);
        if (dtm != null) {
            dtm.setJdbcConnectionHandler(this);
        }
        return dtm;
    }

    public static String createPrimaryKey() {
        if (primaryKey == 0L) {
            primaryKey = System.currentTimeMillis();
        }
        return Long.toString(++primaryKey);
    }
}
