//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.layout;

import com.nexirius.framework.datamodel.DataModel;
import com.nexirius.framework.dataviewer.ViewerFactory;

import javax.swing.*;
import java.awt.*;

/**
 * This class holds a PropertyChanger and uses it to change specific JComponent properties
 * like color font or alignement.
 *
 * @author Marcel Baumann
 */
public class SinglePropertyEntry implements PropertyEntry {



    public static final String PROPERTY_FONT = "Font";
    public static final String PROPERTY_FOREGROUND = "Foreground";
    public static final String PROPERTY_BACKGROUND = "Background";
    public static final String PROPERTY_ALIGNMENT = "Alignment";

    String name;
    String value;
    Object params[];
    PropertyChanger changer = null;

    public SinglePropertyEntry() {
    }

    /**
     * Creates a PropertyChanger SinglePropertyEntry instance
     * <p/>
     * <PRE>
     * protected static MultiplePropertyEntry rightProps = new MultiplePropertyEntry();
     * <p/>
     * private static void init()
     * {
     * if (initDone) {
     * <p/>
     * return;
     * }
     * <p/>
     * initDone = true;
     * <p/>
     * rightProps.append(new SinglePropertyEntry(new MyTablePropertyChanger()));
     * }
     * <p/>
     * class MyTablePropertyChanger extends PropertyChanger
     * {
     * public void change(JComponent comp)
     * {
     * if (comp instanceof JTable) {
     * ((JTable)comp).setGridColor(Color.red);
     * }
     * }
     * <p/>
     * }
     * </PRE>
     */
    public SinglePropertyEntry(PropertyChanger changer) {
        this.changer = changer;
    }

    /**
     * Creates an instance of a PropertyChanger holder
     *
     * @param name  property name (PROPERTY_FONT, PROPERTY_FOREGROUND, PROPERTY_BACKGROUND, PROPERTY_ALIGNMENT)
     * @param value property value
     */
    public SinglePropertyEntry(String name, String value) {
        this.name = name;
        this.value = value;
    }


    /**
     * Creates an instance of a PropertyChanger holder
     *
     * @param methodName Name of the actual JComponent method (e.g. "setBorder")
     * @param params     array of parameters to the JComponent method
     */
    public SinglePropertyEntry(String methodName, Object params[]) {
        this.name = methodName;
        this.params = params;
    }

    /**
     * Creates an instance of a PropertyChanger holder
     *
     * @param methodName Name of the actual JComponent method (e.g. "setBorder")
     * @param param0     parameter to the JComponent method
     */
    public SinglePropertyEntry(String methodName, Object param0) {
        this.name = methodName;
        this.params = new Object[1];
        params[0] = param0;
    }

    /**
     * Creates an instance of a PropertyChanger holder
     *
     * @param methodName Name of the actual JComponent method (e.g. "setBorder")
     * @param param0     parameter to the JComponent method
     * @param param1     parameter to the JComponent method
     */
    public SinglePropertyEntry(String methodName, Object param0, Object param1) {
        this.name = methodName;
        this.params = new Object[2];
        params[0] = param0;
        params[1] = param1;
    }

    /**
     * Creates an instance of a PropertyChanger holder
     *
     * @param methodName Name of the actual JComponent method (e.g. "setBorder")
     * @param param0     parameter to the JComponent method
     * @param param1     parameter to the JComponent method
     * @param param2     parameter to the JComponent method
     */
    public SinglePropertyEntry(String methodName, Object param0, Object param1, Object param2) {
        this.name = methodName;
        this.params = new Object[3];
        params[0] = param0;
        params[1] = param1;
        params[2] = param2;
    }

    /**
     * Creates an instance of a PropertyChanger holder
     *
     * @param methodName Name of the actual JComponent method (e.g. "setBorder")
     * @param param0     parameter to the JComponent method
     * @param param1     parameter to the JComponent method
     * @param param2     parameter to the JComponent method
     * @param param3     parameter to the JComponent method
     */
    public SinglePropertyEntry(String methodName, Object param0, Object param1, Object param2, Object param3) {
        this.name = methodName;
        this.params = new Object[4];
        params[0] = param0;
        params[1] = param1;
        params[2] = param2;
        params[3] = param3;
    }

    /**
     * Change the given component with the appropriate PropertyChanger
     */
    public void change(JComponent comp, ViewerFactory factory)
            throws Exception {
        getChanger(factory).change(comp);
    }

    public void childValueChanged(DataModel child) {
        changer = null;
    }

    /**
     * Creates the appropriate PropertyChanger (if not already defined) and returns it.
     */
    public PropertyChanger getChanger(ViewerFactory factory)
            throws Exception {
        if (changer == null) {
            changer = createPropertyChanger(name, value, params, factory);
        }

        return changer;
    }

    /**
     * Creates a PropertyChanger instance from a name and a value
     */
    public static PropertyChanger createPropertyChanger(String name, String value, Object params[], ViewerFactory factory) {
        if (PROPERTY_FONT.equals(name)) {

            return new FontChanger(factory.getFont(value));
        } else if (PROPERTY_FOREGROUND.equals(name)) {

            return new ForegroundChanger(factory.getColor(value));
        } else if (PROPERTY_BACKGROUND.equals(name)) {

            return new BackgroundChanger(factory.getColor(value));
        } else if (PROPERTY_ALIGNMENT.equals(name)) {

            return new AlignmentChanger(value);
        }

        return new DefaultPropertyChanger(name, params);
    }

    public static SinglePropertyEntry alignLeft() {
        return new SinglePropertyEntry(PROPERTY_ALIGNMENT, AlignmentChanger.LEFT);
    }

    public static SinglePropertyEntry alignCenter() {
        return new SinglePropertyEntry(PROPERTY_ALIGNMENT, AlignmentChanger.CENTER);
    }

    public static SinglePropertyEntry alignRight() {
        return new SinglePropertyEntry(PROPERTY_ALIGNMENT, AlignmentChanger.RIGHT);
    }

    public static SinglePropertyEntry font(String font) {
        return new SinglePropertyEntry(PROPERTY_FONT, font);
    }

    public static SinglePropertyEntry foreground(String foreground) {
        return new SinglePropertyEntry(PROPERTY_FONT, foreground);
    }

    public static SinglePropertyEntry background(String background) {
        return new SinglePropertyEntry(PROPERTY_FONT, background);
    }

    public SinglePropertyEntry preferredSize(int width, int height) {
        return new SinglePropertyEntry("setPreferredSize", new Dimension(width, height));
    }
}
