//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 *
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 *
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 *
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 *
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.swing;

import com.nexirius.framework.datamodel.DataModel;
import com.nexirius.framework.datamodel.Viewable;
import com.nexirius.framework.dataviewer.DataViewer;
import com.nexirius.framework.dataviewer.ViewerCreator;

import java.lang.reflect.Constructor;

/**
 * This class is a factory for SwingViewer instances
 * <PRE>
 * getFactory().getViewerCreatorMap().registerTemplate(MyDataModel.class, new SwingViewerCreator(MySwingViewer.class));
 * <p/>
 * class MyDataModel extends StructModel
 * {
 * //...
 * }
 * <p/>
 * public class MyScrollPane extends JScrollPane implements SwingViewer {
 * <p/>
 * <p/>
 * <p/>
 * DataModel model;
 * <p/>
 * public MyScrollPane()
 * {
 * }
 * <p/>
 * public JComponent createJComponent(ViewerFactory factory) {
 * try {
 * // create a default viewer JComponent which is associated to the given model
 * DataViewer dataViewer = factory.createDefaultViewer(model);
 * JComponent comp = dataViewer.getJComponent();
 * <p/>
 * // put the created JComponent into the JScrollPane (to be the actual viewport view)
 * setViewportView(comp);
 * } catch (Exception ex)
 * {
 * throw new RuntimeException(ex);
 * }
 * <p/>
 * return this;
 * }
 * <p/>
 * public boolean isEditor() {
 * return false;
 * }
 * <p/>
 * public void setDataModel(DataModel model) {
 * this.model = model;
 * }
 * <p/>
 * public DataModel getDataModel() {
 * return model;
 * }
 * }
 * </PRE>
 *
 * @author Marcel Baumann
 */
public class SwingViewerCreator implements ViewerCreator {
    Class<?> swingViewerClass;
    boolean isEditor;
    Object parameter;
    private Object constructorParameter;
    private Constructor<?> constructor;

    public SwingViewerCreator(Class<?> swingViewerClass) {
        this(swingViewerClass, null);
    }

    /**
     * use this viewer creator to create a SwingViewer instance
     *
     * @param swingViewerClass the class of the SwingViewer
     */
    public SwingViewerCreator(Class<?> swingViewerClass, Object constructorParameter) {
        this.swingViewerClass = swingViewerClass;
        this.constructorParameter = constructorParameter;

        try {
            if (constructorParameter == null) {
                constructor = swingViewerClass.getConstructor();
            } else {
                constructor = swingViewerClass.getConstructor(constructorParameter.getClass());
            }
        } catch (NoSuchMethodException ex) {
            throw new RuntimeException(swingViewerClass + " does not have constructor(" + (constructorParameter == null ? "" : constructorParameter.getClass()) + ") " + SwingViewer.class);
        }

        try {
            SwingViewer viewer = createSwingViewer();
            isEditor = viewer.isEditor();
        } catch (Exception ex) {
            throw new RuntimeException("Cannot create new instance of type " + swingViewerClass + " (" + ex.getMessage() + ")", ex);
        }

    }

    private SwingViewer createSwingViewer() {
        Object viewerObject;

        try {

        if (constructorParameter == null) {
            viewerObject = this.swingViewerClass.newInstance();
        } else {
            viewerObject = constructor.newInstance(constructorParameter);
        }
        } catch (Exception ex) {
            throw new RuntimeException(ex);
        }

        if (!(viewerObject instanceof SwingViewer)) {
            throw new RuntimeException(swingViewerClass + " does not implement " + SwingViewer.class);
        }

        return (SwingViewer) viewerObject;
    }

    @Override
    public boolean isEditor() {
        return isEditor;
    }

    /**
     * creates an instance of type SwingDataViewer using an instance of type SwingViewer as constructor parameter
     */
    @Override
    public DataViewer createViewer(Viewable model) {
        if (model == null) {
            throw new RuntimeException("model must not be null");
        }

        if (!(model instanceof DataModel)) {
            throw new RuntimeException("need a type DataModel to create a SwingDataViewer but have " + model.getClass());
        }

        try {
            SwingViewer viewer = createSwingViewer();
            viewer.setDataModel((DataModel) model);

            SwingDataViewer dataViewer = new SwingDataViewer(viewer);

            return dataViewer;
        } catch (Exception ex) {
            throw new RuntimeException("Cannot create SwingDataViewer " + swingViewerClass + " (" + ex.getMessage() + ")", ex);
        }
    }
}
