//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 *
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 *
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 *
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 *
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.framework.textselector;

import com.nexirius.framework.application.DialogManager;
import com.nexirius.framework.datamodel.*;
import com.nexirius.util.CopyPairs;
import com.nexirius.util.StringVector;

/**
 * This is the default implementation to handle a SelectorModel
 * It containes a StringModel (field name: "SelectorModelText") and a ComboBoxModel (field name: "SelectorModelSelector")
 * as well as a command (name "selectorModelAdd") which is used to add the currrent field value to the selectionArray (SimpleArray which
 * is displayed in the ComboBoxModel)
 * <PRE>
 * //Example:
 * String arr[] = {"select1", "select2"};
 * SimpleArrayModel titleArray = new SimpleArrayModel(arr);
 * TextSelectorManager.getInstance().addTextSelector(new DefaultTextSelector("mySelector", titleArray));
 * </PRE>
 */
public class DefaultTextSelector extends StructModel implements TextSelector {



    public static int maxCompletionList = 10;
    StringModel text;
    ComboBoxModel selector;
    StringVector selectionVector = new StringVector();
    DefaultDataModelCommand addCommand;
    int minCharsForPopup = 2;

    public DefaultTextSelector(String name, String selection[]) {
        super(name);

        selectionVector.sortInsert(selection);

        if (selectionVector.size() == 0) {
            selectionVector.append("");
        }

        init();
    }

    public void setSelection(String selection[]) {
        selectionVector.sortInsert(selection);

        updateSelectionVector();
    }

    public StringVector getSelectionVector() {
        return selectionVector;
    }

    public void updateSelectionVector() {
        if (selectionVector.size() == 0) {
            selectionVector.append("");
        }

        selector.getValueArray().setArray((Object[])selectionVector.getArray());
    }

    /**
     * get the name which is used to reference this unique instance via TextSelectorManager
     */
    @Override
    public String getName() {
        return getFieldName();
    }

    /**
     * registerTemplate new text entry (must call setText(text))
     */
    @Override
    public synchronized void addText(String text) {
        setText(text);
        selectorModelAdd();
    }

    /**
     * get a possible completion for a given string
     *
     * @return the actual possible completion or null if none of the entries starts with the given string
     */
    @Override
    public synchronized String getCompletionFor(String s) {
        if (s == null) {

            return null;
        }

        setText(s);

        if (selector.getText() == null || !selector.getText().toUpperCase().startsWith(s.toUpperCase())) {

            return null;
        }

        return selector.getText();
    }

    /**
     * When a certain amount of characters is reached, then the popup window with maching entries is activated
     *
     * @return the threshhold
     */
    @Override
    public int getMinCharsForPopup() {
        return minCharsForPopup;
    }

    /**
     * get all possible completions for a given string
     *
     * @param s      the string which will be tested against
     * @param result the resulting list of strings will be appended to this list
     * @return true if there are more than the returned number of possible completions
     */
    @Override
    public boolean getCompletionListFor(String s, StringVector result) {
        if (s == null) {

            return false;
        }

        setText(s);
        int i = 0;

        for (String str = selectionVector.firstItem(); str != null; str = selectionVector.nextItem()) {

            if (str.toUpperCase().startsWith(s.toUpperCase())) {

                if (i >= maxCompletionList) {

                    return true;
                }

                result.sortInsert(str);
                ++i;
            }
        }

        return false;
    }

    /**
     * access the text which has been selected
     */
    @Override
    public String getText() {
        return text.getText();
    }

    /**
     * set the text which will be selected
     */
    @Override
    public void setText(String s) {
        text.setText(s);
    }

    /**
     * popup a dialog which is used to edit the actual text which will be selected
     *
     * @return true if selection has been performed
     */
    @Override
    public boolean popup() {
        DialogManager.getPopupEditorAdaptor().noDuplicatePopupEdit(this, null, null);

        return true;
    }

    private void init() {
        append(selector = new ComboBoxModel(0, new SimpleArrayModel((Object[])selectionVector.getArray()), "SelectorModelSelector"));
        append(text = new StringModel("", "SelectorModelText"));

        text.addDataModelListener(new TextListener());
        selector.addDataModelListener(new SelectorListener());

        addCommand = new DefaultDataModelCommand("selectorModelAdd");

        addCommand.setSynchronous(true);
        appendMethod(addCommand);
    }

    public synchronized void update() {
        String s = text.getText();


        if (s.length() >= 2) {
            addCommand.setEnabled(!s.equals(selector.getText()));
        } else {
            addCommand.setEnabled(false);
        }
    }

    public synchronized void selectorModelAdd() {
        String newElement = text.getText();

        if (newElement.length() == 0 || text.isEmpty()) {

            return;
        }

        if (newElement.equals(getCompletionFor(newElement))) {

            return;
        }

        selectionVector.sortInsert(newElement);

        try {
            selector.getValueArray().setArray((Object[])selectionVector.getArray());
            setText(newElement);
        } catch (Exception ex) {
            throw new RuntimeException(ex);
        }

        update();
    }

    @Override
    public synchronized DataModel duplicate(DataModel instance, CopyPairs copyPairs) {
        if (instance == null) {
            instance = new DefaultTextSelector(getFieldName(), selectionVector.getArray());
        }

        return super.duplicate(instance, copyPairs);
    }

    public class SelectorListener extends DataModelAdaptor {
        @Override
        public void dataModelChangeValue(DataModelEvent ev) {
            try {
                text.setText(selector.getText());
            } catch (Exception ex) {
                throw new RuntimeException(ex);
            }
        }
    }

    public class TextListener extends DataModelAdaptor {
        @Override
        public void dataModelChangeValue(DataModelEvent ev) {
            String s = text.getText();

            try {
                if (s.length() > 0) {
                    selector.setText(s);
                }
            } catch (Exception ex) {
                throw new RuntimeException(ex);
            }

            update();
        }

        @Override
        public void dataModelEdit(DataModelEvent ev) {
            String s = text.getText();

            if (ev.getId() == DataModelEvent.VALUE_EDIT) {
                try {
                    if (s.length() > 0) {
                        selector.setText(s);
                    }
                } catch (Exception ex) {
                    throw new RuntimeException(ex);
                }
                update();
            } else if (ev.getId() == DataModelEvent.START_EDIT) {
                if (s.length() > 0) {
                    try {
                        selector.setText(s);
                    } catch (Exception ex) {
                        throw new RuntimeException(ex);
                    }
                    update();
                }
            } else if (ev.getId() == DataModelEvent.FINISH_EDIT) {
                update();
            }
        }
    }

    /**
     * @param i
     */
    @Override
    public void setMinCharsForPopup(int i) {
        minCharsForPopup = i;
    }

}
