//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 *
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 *
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 *
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 *
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.util;

import java.lang.ref.SoftReference;
import java.util.ArrayList;
import java.util.Comparator;

/**
 * this class handles a list of event listeners. <BR>
 * Use it like this: <BR>
 * <p/>
 * <pre>
 * <p/>
 * interface AnyEvent
 * {
 * }
 * <p/>
 * interface AnyListener
 * {
 * public void function(AnyEvent event);
 * }
 * <p/>
 * class Any {
 * Object anyListenerList = null;
 * <p/>
 * public void addAnyListener(AnyListener listener)
 * {
 * anyListenerList = EventMulticaster.addEventListener(anyListenerList, listener);
 * }
 * <p/>
 * public void addSoftAnyListener(AnyListener listener)
 * {
 * anyListenerList = EventMulticaster.addEventListener(anyListenerList, new SoftReference(listener));
 * }
 * <p/>
 * public void removeAnyListener(AnyListener listener)
 * {
 * anyListenerList = EventMulticaster.removeEventListener(anyListenerList, listener);
 * }
 * <p/>
 * public void callAnyListenerFunction(AnyEvent event)
 * {
 * Object listeners[] = EventMulticaster.getArray(anyListenerList);
 * boolean needGarbageCollect = false;
 * <p/>
 * for (int i = 0; listeners != null && i < listeners.length; ++i) {
 * AnyListener listener = (AnyListener)listeners[i];
 * <p/>
 * if (listener == null) {
 * needGarbageCollect = true;
 * } else {
 * listener.function(event);
 * }
 * }
 * <p/>
 * if (needGarbageCollect) {
 * anyListenerList = EventMulticaster.garbageCollect(anyListenerList);
 * }
 * }
 * }
 * <p/>
 * </pre>
 */
public class EventMulticaster extends ArrayList<Object> {

    private static final long serialVersionUID = 3887820164735415614L;

    private static EqualListenerComparer equalListenerComparer = new EqualListenerComparer();
    private static SimilarListenerComparer similarListenerComparer = new SimilarListenerComparer();

    private EventMulticaster(Object l1, Object l2) {
        super();
        add(l1);
        add(l2);
    }

    /**
     * Adds a lister to an existing listener or to an existing listener list!
     * The list is created as necessary. This happens if a second listener is
     * added to a first one.
     * @param listenerList The existing listener or existing listener list.
     * May also be null.
     * @param newListener The listener to add to the list.
     * @return Either the listener itself or a listener list containing all listeners.
     */
    public static Object addEventListener(Object listenerList, Object newListener) {
        if (newListener == null) {
        	// Nothing to add ==> Return the original list.
            return listenerList;
        }

        if (listenerList == null) {
        	// No list available yet ==> Directly return the one and only listener
            return newListener;
        } else if (listenerList instanceof EventMulticaster) {
        	// The listerList is already a EventMulticaster
        	// ==> Add the new lister to its lister list.
            ((EventMulticaster) listenerList).add(newListener);

            return listenerList;
        }

        // Create a new EventMulticaster out of the two listeners
        return new EventMulticaster(listenerList, newListener);
    }

    /**
     * Removes a similar listener from a listener list.
     * @param listenerList The existing listener list.
     * @param listener The listener to remove.
     * @return The listener list without the listener
     */
    public static Object removeSimilarEventListener(Object listenerList, Object listener) {
        listenerList = garbageCollect(listenerList);
        return (removeEventListener(listenerList, listener, similarListenerComparer));
    }

    /**
     * Removes a listener from a listener list.
     * @param listenerList The existing listener list.
     * @param listener The listener to remove.
     * @return The listener list without the listener
     */
    public static Object removeEventListener(Object listenerList, Object listener) {
        return (removeEventListener(listenerList, listener, equalListenerComparer));
    }

    /**
     * Removes a listener from a listener list.
     * @param listenerList The existing listener list.
     * @param listener The listener to remove.
     * @return The listener list without the listener
     */
    protected static Object removeEventListener(Object listenerList, Object listener, Comparator<Object> comparator) {
        if (listener == null) {
        	// Nothing to remove ==> Return the original list.
            return listenerList;
        }

        if (listenerList == null) {
        	// Cannot remove the listener from the empty list ==> Return that empty list.
            return null;
        }

        synchronized (listenerList) {
            if (listenerList instanceof EventMulticaster) {
            	// The listener list really is a list!
                EventMulticaster v = ((EventMulticaster) listenerList);

                try {
                    for (int i = 0; i < v.size(); i++)
                    {
                        Object l = v.get(i);
                        if (comparator.compare(l, listener)  == 0)
                        {
                            v.remove(i);
                            break;
                        }
                    }
                } catch (Exception ex) {
                    throw new RuntimeException(ex);
                }

                if (v.size() == 1) {
                	// Only one element left
                	// ==> Return that element instead of the list with the single element.
                    return v.get(0);
                }

                if (v.size() == 0) {
                    throw new RuntimeException("No more elements in list");
                }

                return v;
            }

            // No EventMulticaster. Probably it is a SoftReference
            if (comparator.compare(listenerList, listener) == 0) {
                return null;
            }
            return listenerList;
        }
    }

    /**
     * Creates and returns an Array of listeners.
     * @param listenerList
     * @return
     */
    public static Object[] getArray(Object listenerList) {
        Object ret[] = null;

        if (listenerList == null) {

            return ret;
        } else if (listenerList instanceof EventMulticaster) {
            synchronized (listenerList) {
                EventMulticaster v = ((EventMulticaster) listenerList);
                ret = new Object[v.size()];
                int index = 0;

                for (Object listener : v) {
                    if (listener instanceof SoftReference) {
                        ret[index++] = ((SoftReference<?>) listener).get();
                    } else {
                        ret[index++] = listener;
                    }
                }
            }
        } else {
            // remove unneeded (empty) soft references
            if (listenerList instanceof SoftReference && ((SoftReference<?>) listenerList).get() == null) {
                listenerList = null;

                return ret;
            }

            ret = new Object[1];

            if (listenerList instanceof SoftReference) {
                ret[0] = ((SoftReference<?>) listenerList).get();
            } else {
                ret[0] = listenerList;
            }
        }

        return ret;
    }

    /**
     * remove all null soft reference entries from the listener list and return the new listener list
     *
     * @param listenerList the listener list which may contains null soft references
     * @return the listener list which does not contain null soft references
     */
    public static Object garbageCollect(Object listenerList) {
        if (listenerList == null) {
            return null;
        }

        synchronized (listenerList) {
            if (listenerList instanceof EventMulticaster) {
                Object newListenerList = null;
                EventMulticaster v = ((EventMulticaster) listenerList);

                for (Object listener : v) {
                    if (listener instanceof SoftReference) {

                        if (((SoftReference<?>) listener).get() != null) {
                            newListenerList = EventMulticaster.addEventListener(newListenerList, listener);
                        }
                    } else {
                        newListenerList = EventMulticaster.addEventListener(newListenerList, listener);
                    }
                }

                return newListenerList;
            }

            if (listenerList instanceof SoftReference && ((SoftReference<?>) listenerList).get() == null) {
                return null;
            }
        }

        return listenerList;
    }

    /**
     * Compares two Listener instances for equallity. Two listener instances are equal if they
     * are equal using the == operator or if they are a valid SoftReference referring the same object.
     */
    public static class EqualListenerComparer implements  Comparator<Object>
    {
        @Override
        public int compare(Object o1, Object o2)
        {
            if (o1 == o2)
            {
                // Same instances ==> Great
                return 0;
            }

            if (o1 instanceof SoftReference && ((SoftReference<?>)o1).get() == o2)
            {
                // A soft reference to the same instances ==> Also not bad
                return 0;
            }

            if (o2 instanceof SoftReference)
            {
                if (((SoftReference<?>)o2).get() == o1)
                {
                    // A SoftReference to the same instances ==> Also not bad
                    return 0;
                }

                if (o1 instanceof SoftReference && ((SoftReference<?>)o1).get() ==  ((SoftReference<?>)o2).get())
                {
                    // Two SoftReference referring the same object ==> Unusal but acceptable
                    return 0;
                }
            }

            // Return some form of 'unequality'
            int ret = o1.hashCode() - o2.hashCode();
            if (ret == 0)
            {
                // Never ever return 0 at this place!
                ret = 1;
            }
            return (ret);
        }
    }

    /**
     * Compares two Listener instances for similarity. Two listener instances are similar if there
     * target is the same class (and to instance of a class)
     */
    public static class SimilarListenerComparer implements  Comparator<Object>
    {
        @Override
        public int compare(Object o1, Object o2)
        {
            Class<?> c1 = o1 instanceof SoftReference ? ((SoftReference<?>)o1).get().getClass() : o1.getClass();
            Class<?> c2 = o2 instanceof SoftReference ? ((SoftReference<?>)o2).get().getClass() : o2.getClass();
            int ret = c1.getName().compareTo(c2.getName());
            return (ret);
        }
    }

}
