//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.PushbackInputStream;

/**
 * A class that reads from a byte stream and compares a magic number at the end of the stream.
 */
public class MagicNumberInputStream extends PushbackInputStream {



    long charCount = 0;
    long magic = 0;
    long magicCompare = 0;
    boolean finished = false;

    /**
     * Creates a <code>MagicNumberInputStream</code>
     * with a pushback buffer of the specified <code>size</code>,
     * and saves its  argument, the input stream
     * <code>in</code>, for later use. Initially,
     * there is no pushed-back byte  (the field
     * <code>pushBack</code> is initialized to
     * <code>-1</code>).
     *
     * @param in   the input stream from which bytes will be read.
     * @param size the size of the pushback buffer.
     * @throws IllegalArgumentException if size is <= 0
     * @since JDK1.1
     */
    public MagicNumberInputStream(InputStream in, int size) {
        super(in, size);
    }

    /**
     * Creates a <code>MagicNumberInputStream</code>
     * and saves its  argument, the input stream
     * <code>in</code>, for later use. Initially,
     * there is no pushed-back byte  (the field
     * <code>pushBack</code> is initialized to
     * <code>-1</code>).
     *
     * @param in the input stream from which bytes will be read.
     */
    public MagicNumberInputStream(InputStream in) {
        this(in, 1);
    }

    public InputStream getIn() {
        return in;
    }

    /**
     * Reads the next byte of data from this input stream. The value
     * byte is returned as an <code>int</code> in the range
     * <code>0</code> to <code>255</code>. If no byte is available
     * because the end of the stream has been reached, the value
     * <code>-1</code> is returned. This method blocks until input data
     * is available, the end of the stream is detected, or an exception
     * is thrown.
     * <p/>
     * <p> This method returns the most recently pushed-back byte, if there is
     * one, and otherwise calls the <code>read</code> method of its underlying
     * input stream and returns whatever value that method returns.
     *
     * @return the next byte of data, or <code>-1</code> if the end of the
     *         stream has been reached.
     * @throws IOException if an I/O error occurs.
     * @see java.io.InputStream#read()
     */
    public int read()
            throws IOException {
        int ret = super.read();

        if (ret == '@') {

            ret = super.read();

            if (ret != '@') {
                setMagicCompare(ret);

                return -1;
            }
        }

        ++charCount;
        magic += charCount * ret;

        return ret;
    }

    private void setMagicCompare(int c)
            throws IOException {
        StringBuffer b = new StringBuffer(20);

        while (c != '@') {
            b.append((char) c);

            c = super.read();

            if (c == -1) {
                throw new IOException("Can't read magic number (Early end of file detected)");
            }
        }

        try {
            magicCompare = Long.parseLong(b.toString());
        } catch (Exception ex) {
            throw new IOException("Can't read magic number format");
        }

        finished = true;
    }

    /**
     * Not yet supported (throws Exception)
     */
    public int read(byte[] b, int off, int len)
            throws IOException {
        throw new IOException("Method MagicNumberInputStream.read(byte[] b, int off, int len) not supported");
    }

    /**
     * Pushes back a byte by copying it to the front of the pushback buffer.
     * After this method returns, the next byte to be read will have the value
     * <code>(byte)b</code>.
     *
     * @param b the <code>int</code> value whose low-order
     *          byte is to be pushed back.
     * @throws IOException If there is not enough room in the pushback
     *                     buffer for the byte.
     */
    public void unread(int b)
            throws IOException {
        super.unread(b);

        magic -= charCount * b;
        --charCount;
    }

    /**
     * Not yet supported (throws Exception)
     */
    public void unread(byte[] b, int off, int len)
            throws IOException {
        throw new IOException("Method MagicNumberInputStream.unread(byte[] b, int off, int len) not supported");
    }

    /**
     * Not yet supported (throws Exception)
     */
    public void unread(byte[] b)
            throws IOException {
        throw new IOException("Method MagicNumberInputStream.unread(byte[] b) not supported");
    }

    /**
     * Not yet supported (throws Exception)
     */
    public long skip(long n)
            throws IOException {
        throw new IOException("Method MagicNumberInputStream.skip(long n) not supported");
    }

    public boolean checkMagic() {
        if (finished) {

            return magic == magicCompare;
        }

        return true;
    }

    public boolean hasMagic() {
        return finished;
    }

    public long getMagicNumber() {
        return magic;
    }
}
