//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 *
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 *
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 *
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 *
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.util;

import java.io.*;
import java.util.Enumeration;
import java.util.Vector;

public class TextToken extends Object {
    public static final int UNDEFINED = 0;
    public static final int INT = 10001;
    public static final int LONG = 10002;
    public static final int DOUBLE = 10003;
    public static final int IDENTIFIER = 10004;
    public static final int STRING = 10005;
    public static final int CHAR = 10006;
    public static final int TAB = 10007; // \t
    public static final int FORM_FEED = 10008; // \f
    public static final int NEW_LINE = 10009; // \n
    static StringBuffer one_char = new StringBuffer("x");

    // if != null using this charset to translate string constants into strings
    public static String CHARSET;

    public static boolean ignoreSpaces = true;

    int tokenType;
    String stringValue;	// holds identifier or string
    long integerValue;		// holds char, int, long
    double doubleValue;
    static int number_bytes = 0;
    static int number_lines = 0;

    /**
     * Creates a token which consists of the specified string which represents the given id
     */
    public TextToken(String token, int id) {
        stringValue = token;
        tokenType = id;
    }

    /**
     * Creates a STRING token. There are no restrictions for the contents or length of a string token.
     * The STRING tokens are written in " characters @see ##quoteString
     */
    public TextToken(String s) {
        tokenType = STRING;
        stringValue = s;
    }

    /**
     * Creates a CHAR token
     */
    public TextToken(char c) {
        tokenType = CHAR;
        integerValue = (long) c;
    }

    /**
     * Creates a INT token
     */
    public TextToken(int i) {
        tokenType = INT;
        integerValue = (long) i;
    }

    /**
     * Creates a LONG token. Long integers are written as integer with a trailing L.
     */
    public TextToken(long l) {
        tokenType = LONG;
        integerValue = l;
    }

    /**
     * Creates a DOUBLE token.
     */
    public TextToken(double d) {
        tokenType = DOUBLE;
        doubleValue = d;
    }

    public static void setCharset(String charset) {
        CHARSET = charset;
    }

    /**
     * Check whether the token is an INT LONG or DOUBLE
     *
     * @return true if the token is a numeric value
     */
    public boolean isNumber() {
        return tokenType == INT || tokenType == LONG || tokenType == DOUBLE;
    }

    public boolean isInt() {
        return tokenType == INT;
    }

    public boolean isLong() {
        return tokenType == LONG;
    }

    public boolean isDouble() {
        return tokenType == DOUBLE;
    }

    public boolean isIdentifier() {
        return tokenType == IDENTIFIER;
    }

    /**
     * Check whether the given string is equal to the IDENTIFIER token.
     *
     * @return true if the token is the identifier id.
     */
    public boolean isIdentifier(String id) {
        if (isIdentifier()) {
            return id.equals(stringValue);
        }

        return false;
    }

    public boolean isString() {
        return tokenType == STRING;
    }

    public boolean isChar() {
        return tokenType == CHAR;
    }

    /**
     * Check wheter this token is the given character
     *
     * @return true if the token is of type CHAR and represents the character c.
     */
    public boolean isChar(char c) {
        if (isChar()) {
            return c == getChar();
        }

        return false;
    }

    /**
     * Is this token a space character
     *
     * @return true if the token is a space tab new line or form feed character
     */
    public boolean isSpace() {
        return tokenType == TAB || tokenType == NEW_LINE || tokenType == FORM_FEED || isChar(' ') || isChar('\r');
    }

    /**
     * Get the type of the token (UNDEFINED, INT, LONG, IDENTIFIER, STRING ...)
     */
    public int getId() {
        return tokenType;
    }

    /**
     * Get the internal string representation (only makes sence for STRING and IDENTIFIER)
     */
    public String getString() {
        return stringValue;
    }

    /**
     * Get the integer (only makes sense if the type is LONG or INT or CHAR)
     */
    public int getInt() {
        return (int) integerValue;
    }

    /**
     * Get the long integer (only makes sense if the type is LONG or INT or CHAR)
     */
    public long getLong() {
        return integerValue;
    }

    /**
     * Get the double value (only makes sense if the type DOUBLE)
     */
    public double getDouble() {
        return doubleValue;
    }

    /**
     * Get the character value (only makes sense if the type is LONG or INT or CHAR)
     */
    public char getChar() {
        return (char) integerValue;
    }

    /**
     * @return the number of bytes that where read to produce the last token read
     */
    public static int getBytes() {
        return number_bytes;
    }

    /**
     * @return the number of lines that where read to produce the last token read
     */
    public static int getLines() {
        return number_lines;
    }

    /**
     * Get the next token of the specified stream
     *
     * @return the next token or null if the last token has been returned already (or on error)
     */
    public static TextToken nextToken(PushbackInputStream in)
            throws IOException {
        TextToken ret = null;

        for (ret = nextTokenIncludingSpaces(in); ret != null; ret = nextTokenIncludingSpaces(in)) {
            if (!ret.isSpace()) {
                break;
            }
        }

        return ret;
    }

    public static void skipLine(PushbackInputStream in)
            throws IOException {
        int c;

        while (true) {
            c = in.read();

            switch (c) {
                case -1:
                    return;

                case '\n':
                    ++number_lines;
                    return;

                default:
                    ++number_bytes;
            }
        }
    }

    public static TextToken nextTokenIncludingSpaces(PushbackInputStream in)
            throws IOException {
        int c;

        c = in.read();
        number_bytes = 1;
        number_lines = 0;

        switch (c) {
            case -1:
                return null;

            case '\"':
                return readString(in);

            case '\t':
                return new TextToken(null, TAB);

            case '\n':
                return new TextToken(null, NEW_LINE);

            case '\f':
                return new TextToken(null, FORM_FEED);

            default:
                break;
        }

        if (Character.isDigit((char) c) || c == '-') {

            return readNumber(in, c);
        }

        if (Character.isJavaIdentifierStart((char) c)) {
            return readIdentifier(in, c);
        }

        return new TextToken((char) c);
    }

    public void writeTo(OutputStream out)
            throws Exception {
        out.write(toString().getBytes());
    }

    public static PushbackInputStream createPushbackInputStream(File file)
            throws FileNotFoundException {
        return new PushbackInputStream(new FileInputStream(file));
    }

    public static PushbackInputStream createPushbackInputStream(String s) {
        return new PushbackInputStream(new ByteArrayInputStream(s.getBytes()));
    }

    /**
     * This function scans the given string and returns a vector containing all the tokens
     * which are in the string.
     */
    public static TextToken[] getTokensFrom(String s) {
        if (s == null) return new TextToken[0];
        Vector vect = new Vector();
        PushbackInputStream is = new PushbackInputStream(new ByteArrayInputStream(s.getBytes()));
        TextToken t = null;

        while (true) {
            try {
                t = TextToken.nextToken(is);

                if (t == null) {
                    break;
                }

                vect.addElement(t);
            } catch (Exception ex) {
                break;
            }
        }

        if (vect.size() == 0) {
            return new TextToken[0];
        }

        TextToken ret[] = new TextToken[vect.size()];
        Enumeration e = vect.elements();
        int i = 0;

        while (e.hasMoreElements()) {
            ret[i++] = (TextToken) e.nextElement();
        }

        return ret;
    }

    /**
     * @return a portable string representation of the token.
     */
    public String toString() {
        switch (tokenType) {
            case INT:
                return Integer.toString(getInt());

            case LONG:
                StringBuffer b = new StringBuffer(Long.toString(getLong()));
                b.append('L');
                return new String(b);

            case DOUBLE:
                return Double.toString(getDouble());

            case IDENTIFIER:
                return getString();

            case STRING:
                return quoteString();

            case CHAR:
                one_char.setCharAt(0, getChar());
                return new String(one_char);

            case TAB:
                one_char.setCharAt(0, '\t');
                return new String(one_char);

            case NEW_LINE:
                one_char.setCharAt(0, '\n');
                return new String(one_char);

            case FORM_FEED:
                one_char.setCharAt(0, '\f');
                return new String(one_char);
        }

        return stringValue;
    }

    public static String quoteString(String str) {
        return quoteString(str, true);
    }

    public static String quoteString(String str, boolean withQuotationMarks) {
        TextToken tok = new TextToken(str);
        return tok.quoteString(withQuotationMarks);
    }

    public String quoteString() {
        return quoteString(true);
    }

    /**
     * @return a quoted (") String where " is replaced by \" and \ is replaced by \\
     *         unicode chars are stored as \uFFFF where FFFF is a hex integer
     */
    public String quoteString(boolean withQuotationMarks) {
        if (stringValue == null) {
            return "\"\"";
        }

        char c[] = new char[stringValue.length()];
        StringBuffer buf = new StringBuffer(6 * c.length + 2); // the max length if all chars are unicode chars
        int i = 0;

        stringValue.getChars(0, c.length, c, 0);

        if (withQuotationMarks) {
            buf.append('\"');
        }

        for (i = 0; i < c.length; ++i) {
            switch (c[i]) {
                case '\"':
                case '\\':
                    buf.append('\\');
                    break;
            }
            if (c[i] < '\u007F' && c[i] > '\u0000') {
                buf.append(c[i]);
            } else if (c[i] == '\u0000') {
                buf.append("\\u0000");
            } else {
                if (c[i] <= '\u00FF') {
                    buf.append("\\u00");
                } else if (c[i] <= '\u0FFF') {
                    buf.append("\\u0");
                } else {
                    buf.append("\\u");
                }

                buf.append(Integer.toHexString((int) c[i]));
            }
        }

        if (withQuotationMarks) {
            buf.append('\"');
        }

        return new String(buf);
    }

    private static TextToken readString(final PushbackInputStream in)
            throws IOException {

        ICharacterContainer characterContainer;

        if (CHARSET != null) {
            final ByteArrayOutputStream out = new ByteArrayOutputStream(10240);

            characterContainer = new ICharacterContainer() {
                public void append(char c) {
                    out.write(c);
                }

                public void append(String s) {
                    try {
                        out.write(s.getBytes(CHARSET));
                    } catch (Exception e) {
                        throw new RuntimeException(e);
                    }
                }

                public String getString() {
                    try {
                        return out.toString(CHARSET);
                    } catch (UnsupportedEncodingException e) {
                        throw new RuntimeException(e);
                    }
                }
            };
        } else {
            final StringBuffer buf = new StringBuffer();

            characterContainer = new ICharacterContainer() {
                public void append(char c) {
                    buf.append(c);
                }

                public void append(String s) {
                    buf.append(s);
                }

                public String getString() {
                    return buf.toString();
                }
            };
        }

        int c;
        int state = 0;
        StringBuffer unicode = new StringBuffer("0000");

        while (true) {
            c = in.read();
            ++number_bytes;

            if (c == -1) {
                throw new IOException("Unexpected end of file inside a string constant");
            }

            state = parseCharFromStringConstant(characterContainer, (char)c, state, unicode);

            if (state == 6) {

                return new TextToken(characterContainer.getString(), STRING);
            }
        }
        // never reached
    }

    private static int parseCharFromStringConstant(ICharacterContainer buf, char c, int state, StringBuffer unicode) {
        switch (state) {
            case 0: // no escape
                switch (c) {
                    case '\\':
                        state = 1;
                        break;

                    case '\"':
                        state = 6; // done
                        break;

                    case '\n':
                        ++number_lines;
                        // FALLTHROUGHT
                    default:
                        buf.append((char) c);
                }
                break;

            case 1: // after escape
                switch (c) {
                    case '\\':
                        state = 0;
                        buf.append('\\');
                        break;

                    case '\"':
                        state = 0;
                        buf.append('\"');
                        break;

                    case 'u':
                        state = 2;
                        break;

                    default:
                        state = 0;
                        buf.append(c); // unexpected escape character
                }
                break;

            case 2: // first unicode character
            case 3: // second unicode character
            case 4: // third unicode character
            case 5: // forth unicode character
                unicode.setCharAt(state - 2, (char) c);
                if (state < 5) {
                    ++state;
                } else {
                    // append unicode character

                    try {
                        buf.append((char) Integer.parseInt(new String(unicode), 16));
                    } catch (Exception ex) {
                        buf.append("\\u" + unicode);
                    }
                    state = 0;
                }

                break;
        }

        return state;
    }

    private static TextToken readNumber(PushbackInputStream in, int first_char)
            throws IOException {
        boolean is_negative = false;
        boolean is_double = false;
        boolean is_long = false;
        int state = 0;
        int c = first_char;
        StringBuffer number = new StringBuffer();

        while (state != -1) {

            switch (state) {
                case 0: // beginning
                    switch (c) {
                        case '-':
                            is_negative = true;
                            number.append('-');
                            break;

                        case '.':
                            state = 2; // after comma
                            number.append('.');
                            break;

                        default:
                            if (Character.isDigit((char) c)) {
                                number.append((char) c);
                                state = 1;
                            } else {
                                in.unread(c);
                                --number_bytes;
                                return new TextToken((char) first_char);
                            }
                    }
                    break;

                case 1: // after first number
                    if (Character.isDigit((char) c)) {
                        number.append((char) c);
                    } else if (c == '.') {
                        number.append('.');
                        state = 2; // after comma
                    } else if (c == 'l' || c == 'L') {
                        is_long = true;
                        state = -1;
                    } else {
                        in.unread(c);
                        --number_bytes;
                        state = -1;
                    }
                    break;

                case 2: // after comma
                    is_double = true;

                    if (Character.isDigit((char) c)) {
                        number.append((char) c);
                    } else {
                        in.unread(c);
                        --number_bytes;
                        state = -1;
                    }
                    break;
            }

            if (state != -1) {
                c = in.read();
                ++number_bytes;

                if (c == -1) {
                    break;
                }
            }
        }

        String s = new String(number);

        try {
            if (is_double) {
                Double d = Double.valueOf(s);

                return new TextToken(d.doubleValue());
            } else if (is_long) {
                return new TextToken(Long.parseLong(s));
            }

            long l = Long.parseLong(s);

            if (l > Integer.MAX_VALUE || l < Integer.MIN_VALUE) {
                return new TextToken(l);
            }

            return new TextToken((int)l);
        } catch (NumberFormatException ex) {
            throw new IOException("Can't read number " + s);
        }
    }

    private static TextToken readIdentifier(PushbackInputStream in, int first_char)
            throws IOException {
        int c = first_char;
        StringBuffer buf = new StringBuffer();

        while (c != -1) {
            if (Character.isJavaIdentifierPart((char) c)) {
                buf.append((char) c);
            } else {
                in.unread(c);
                --number_bytes;
                break;
            }

            c = in.read();
            ++number_bytes;
        }

        return new TextToken(new String(buf), IDENTIFIER);
    }

    /**
     * @return the token type as a string (mainly for debug)
     */
    public String getTypeAsString() {
        switch (tokenType) {
            case INT:
                return new String("INT");

            case LONG:
                return new String("LONG");

            case DOUBLE:
                return new String("DOUBLE");

            case IDENTIFIER:
                return new String("IDENTIFIER");

            case STRING:
                return new String("STRING");

            case CHAR:
                return new String("CHAR");

            case TAB:
                return new String("TAB");

            case NEW_LINE:
                return new String("NEW_LINE");

            case FORM_FEED:
                return new String("FORM_FEED");

            default:
                return new String("unknown " + tokenType);
        }
    }

    /**
     * writes the type and the token to the standard output
     */
    public void debugPrint() {
        System.out.println(getTypeAsString() + ": " + toString());
    }

    /**
     * returns the type and value as String
     */
    public String debugString() {
        return getTypeAsString() + ": " + toString();
    }

    interface ICharacterContainer {
        public void append(char c);
        public void append(String s);
        public String getString();
    }

//    public static void main(String[] argv) throws IOException {
//        TextToken.setCharset("UTF-8");
//        TextToken[] tokens = TextToken.getTokensFrom("Dies ist ein \"Test\" 12345 \"‰äöü\u2030a\"");
//
//        for (int i = 0; i < tokens.length; i++) {
//            TextToken token = tokens[i];
//
//            System.out.println("token = " + token);
//
//            System.out.println("token.getString() = " + token.getString());
//        }
//    }
}
