//{HEADER
/**
 * This class is part of jnex 'Nexirius Application Framework for Java'
 *	
 * Copyright (C) Nexirius GmbH, CH-4450 Sissach, Switzerland (www.nexirius.ch)
 * 
 * <p>This library is free software; you can redistribute it and/or<br>
 * modify it under the terms of the GNU Lesser General Public<br>
 * License as published by the Free Software Foundation; either<br>
 * version 2.1 of the License, or (at your option) any later version.</p>
 * 
 * <p>This library is distributed in the hope that it will be useful,<br>
 * but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU<br>
 * Lesser General Public License for more details.</p>
 * 
 * <p>You should have received a copy of the GNU Lesser General Public<br>
 * License along with this library; if not, write to the Free Software<br>
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA</p>
 * </blockquote>
 * 
 * <p>
 * Nexirius GmbH, hereby disclaims all copyright interest in<br>
 * the library jnex' 'Nexirius Application Framework for Java' written<br>
 * by Marcel Baumann.</p>
 */
//}HEADER
package com.nexirius.util.statemachine;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * The state is a container for transistions. Each State can have an own EventSource.
 * The default transition (holds a null event) is targeted on each event which is not
 * mapped to one of the transitions.
 */
public class State implements Serializable {



    String name;
    boolean isEnd = false;
    EventSource eventSource = null;
    Transition defaultTransition = null;
    Hashtable transitions = new Hashtable();

    /**
     * This constructor can be used to define the end state
     */
    public State() {
        isEnd = true;
    }

    /**
     * Define a state
     *
     * @param name The name of the state
     */
    public State(String name) {
        this.name = name;
    }

    /**
     * Definie a state which holds its own event source instance
     *
     * @param name The name of the state
     * @param es   The EventSource instance which is used to generate events to be handled on this state (may be null
     *             if the state machine gets a event source parameter while stepping from one state to another)
     */
    public State(String name, EventSource es) {
        this.name = name;
        this.eventSource = es;
    }

    /**
     * Define a state with a name which could be the end state
     *
     * @param name  the name of the state
     * @param isEnd When this is true then the state machine stops (an end state does not have transitions)
     */
    public State(String name, boolean isEnd) {
        this.name = name;
        this.isEnd = isEnd;
    }

    /**
     * access the event source
     *
     * @return
     */
    public EventSource getEventSource() {
        return eventSource;
    }

    /**
     * Adds a transition (which holds an event and a target state)
     * There may only one transition be attached to the state with the same event.
     *
     * @param tr the transition
     * @throws RuntimeException if an transition with the same event has already been registered for that state
     */
    public void addTransition(Transition tr) {
        if (tr.isDefault()) {

            if (defaultTransition != null) {
                throw new RuntimeException("Only one default transition allowed on state:" + name);
            }

            defaultTransition = tr;
        } else {
            if (transitions.get(tr.getEvent()) != null) {

                throw new RuntimeException("Only one transition allowed for event (" + tr.getEvent() + ") on state:" + name);
            }

            transitions.put(tr.getEvent(), tr);
        }
    }

    /**
     * access a transition by its event name
     *
     * @param event the name of the transition or null for the default transition
     * @return If event is null or not registered for this state then the default transition is returned
     */
    public Transition getTransition(String event) {
        if (event == null) {

            return defaultTransition;
        }

        Transition ret = (Transition) transitions.get(event);

        if (ret == null) {
            ret = defaultTransition;
        }

        return ret;
    }

    /**
     * access the transition which handles the event and return its target state
     *
     * @param event The event (or null for the default transition)
     * @return The next state in the state machine or null (only if no default transition was registered)
     */
    public State getNextStateForEvent(String event) {
        Transition tr = getTransition(event);

        if (tr == null) {

            return null;
        }

        return tr.getNextState();
    }

    /**
     * same as getNextStateForEvent(null)
     *
     * @return
     */
    public State getNextStateForDefaultEvent() {
        if (defaultTransition == null) {

            return null;
        }

        return defaultTransition.getNextState();
    }

    /**
     * Access an iterator to all registered events for this state
     *
     * @return The Enumeration which holds a String list of all registered events
     */
    public Enumeration getAllAcceptedEvents() {
        return transitions.keys();
    }

    /**
     * Same as getNextStateForEvent
     */
    public State step(String event) {
        Transition tr = getTransition(event);

        if (tr == null) {

            return null;
        }

        return tr.getNextState();
    }

    /**
     * Access the name of the state
     *
     * @return
     */
    public String getName() {
        return name;
    }

    /**
     * Check whether this state is supposed to be the end state of a state machine
     *
     * @return
     */
    public boolean isEnd() {
        return isEnd;
    }
}

